"""
Утилиты для работы с конфигурацией базы данных.

Предоставляет функции для загрузки и вычисления database_config
с учетом всех источников данных (database_config.json, credentials, params).
"""
import os
import json
from pathlib import Path
from typing import Dict, Optional, Any
from ..core.logger import get_module_logger

logger = get_module_logger("database_config_utils")


def get_database_config_path() -> Path:
    """Получить путь к файлу database_config.json"""
    return Path(__file__).parent.parent / "database_config.json"


def load_database_config_from_file() -> Optional[Dict[str, Any]]:
    """
    Загрузить базовую конфигурацию БД из файла database_config.json.
    
    Returns:
        Словарь с конфигурацией БД или None, если файл не найден
    """
    config_path = get_database_config_path()
    try:
        if config_path.exists():
            with open(config_path, 'r', encoding='utf-8') as f:
                config = json.load(f)
            logger.debug(f"Loaded database_config from {config_path}")
            return config
        else:
            logger.warning(f"database_config.json not found at {config_path}")
            return None
    except Exception as e:
        logger.error(f"Failed to load database_config.json: {e}")
        return None


def compute_database_config(
    use_database: bool = True,
    credentials: Optional[Dict[str, Any]] = None,
    params: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    """
    Вычислить полную конфигурацию БД с учетом всех источников.
    
    Логика вычисления:
    1. Загружается базовая конфигурация из database_config.json
    2. Заполняются значения по умолчанию из credentials
    3. Переопределяются значения из params['database'] (если есть)
    
    Args:
        use_database: Включена ли БД (если False, возвращает пустой config)
        credentials: Словарь с credentials (опционально)
        params: Словарь с параметрами конфигурации (опционально)
    
    Returns:
        Полная конфигурация БД в формате:
        {
            "database": {...},
            "database_adapters": {...}
        }
    """
    if not use_database:
        logger.debug("Database disabled, returning empty config")
        return {"database": {}, "database_adapters": {}}
    
    # Загружаем базовую конфигурацию из файла
    database_config = load_database_config_from_file()
    if database_config is None:
        # Если файл не найден, создаем минимальную структуру
        database_config = {"database": {}, "database_adapters": {}}
    
    # Убеждаемся, что есть ключи database и database_adapters
    if "database" not in database_config:
        database_config["database"] = {}
    if "database_adapters" not in database_config:
        database_config["database_adapters"] = {}
    
    # Получаем credentials с значениями по умолчанию
    database_creds = {}
    if credentials:
        database_creds = credentials.get("database", {}) or {}
    
    # Устанавливаем значения по умолчанию для credentials
    database_creds.setdefault("user_name", "postgres")
    database_creds.setdefault("password", "")
    database_creds.setdefault("database_name", "evil_eye_db")
    database_creds.setdefault("host_name", "localhost")
    database_creds.setdefault("port", 5432)
    database_creds.setdefault("admin_user_name", "postgres")
    database_creds.setdefault("admin_password", "")
    
    # Объединяем значения: сначала из файла, потом из credentials
    db_section = database_config["database"]
    db_section.setdefault("user_name", database_creds["user_name"])
    db_section.setdefault("password", database_creds["password"])
    db_section.setdefault("database_name", database_creds["database_name"])
    db_section.setdefault("host_name", database_creds["host_name"])
    db_section.setdefault("port", database_creds["port"])
    db_section.setdefault("admin_user_name", database_creds["admin_user_name"])
    db_section.setdefault("admin_password", database_creds["admin_password"])
    
    # Переопределяем значения из params['database'], если есть
    if params and "database" in params:
        params_db = params["database"]
        if isinstance(params_db, dict):
            # Обновляем только те ключи, которые есть в params
            if "database_name" in params_db:
                db_section["database_name"] = params_db["database_name"]
            if "host_name" in params_db:
                db_section["host_name"] = params_db["host_name"]
            if "port" in params_db:
                db_section["port"] = params_db["port"]
            if "image_dir" in params_db:
                db_section["image_dir"] = params_db["image_dir"]
            if "preview_width" in params_db:
                db_section["preview_width"] = params_db["preview_width"]
            if "preview_height" in params_db:
                db_section["preview_height"] = params_db["preview_height"]
    
    logger.debug(f"Computed database_config with database keys: {list(db_section.keys())}")
    return database_config


def ensure_database_config_complete(database_config: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Убедиться, что database_config содержит все необходимые ключи,
    при этом не перетирая уже заданные пользователем значения.
    
    Args:
        database_config: Существующая конфигурация БД (может быть неполной)
    
    Returns:
        Полная конфигурация БД с гарантированными ключами 'database' и 'database_adapters'
    """
    if not database_config:
        database_config = {}
    
    # Загружаем базовую конфигурацию из файла для ДОПОЛНЕНИЯ
    file_config = load_database_config_from_file()
    
    # Набор ключей, по которым можно распознать "плоскую" секцию database
    db_flat_keys = {
        "user_name",
        "password",
        "database_name",
        "host_name",
        "port",
        "admin_user_name",
        "admin_password",
        "image_dir",
        "preview_width",
        "preview_height",
        "tables",
    }
    
    # 1. Обеспечиваем наличие секции 'database'
    if "database" not in database_config:
        # Возможен сценарий, когда нам передали уже «плоскую» секцию database
        # (c ключами host_name/port и т.п.), но без обёртки 'database'
        if any(k in database_config for k in db_flat_keys):
            # Выделяем часть, похожую на секцию database, и оборачиваем её
            db_section = {k: v for k, v in database_config.items() if k in db_flat_keys}
            other_keys = {k: v for k, v in database_config.items() if k not in db_flat_keys}
            database_config = {
                **other_keys,
                "database": db_section,
            }
            logger.info(
                "Interpreting provided config as flattened 'database' section; "
                "wrapping it into full database_config structure"
            )
        elif file_config and "database" in file_config:
            # Если плоской секции нет, используем дефолты из файла
            database_config["database"] = file_config["database"].copy()
            logger.info("Loaded 'database' section from database_config.json file")
        else:
            database_config["database"] = {}
            logger.warning("'database' section not found, using empty dict")
    
    # 2. Обеспечиваем наличие секции 'database_adapters'
    if "database_adapters" not in database_config:
        if file_config and "database_adapters" in file_config:
            database_config["database_adapters"] = file_config["database_adapters"].copy()
            logger.info("Loaded 'database_adapters' section from database_config.json file")
        else:
            database_config["database_adapters"] = {}
            logger.warning("'database_adapters' section not found, using empty dict")
    
    # 3. Дополняем недостающие ключи внутри 'database' из файла,
    # НЕ перезаписывая уже заданные значения (host, port и т.п.)
    if file_config and "database" in file_config:
        db_section = database_config.get("database") or {}
        file_db_section = file_config.get("database") or {}
        for key, value in file_db_section.items():
            db_section.setdefault(key, value)
        database_config["database"] = db_section
    
    return database_config



#!/usr/bin/env python3
"""
Утилиты для логирования в EvilEye

Предоставляет удобные функции для логирования в различных модулях приложения.
"""

import logging
import functools
import time
from typing import Any, Callable, Optional
from contextlib import contextmanager


def get_module_logger(module_name: str) -> logging.Logger:
    """
    Получить логгер для указанного модуля
    
    Args:
        module_name: Имя модуля (например, __name__)
        
    Returns:
        Логгер для модуля
    """
    return logging.getLogger(f"evileye.{module_name.split('.')[-1]}")


def log_function_call(logger: Optional[logging.Logger] = None, 
                     level: int = logging.DEBUG,
                     log_args: bool = True,
                     log_result: bool = True):
    """
    Декоратор для логирования вызовов функций
    
    Args:
        logger: Логгер для использования (если None, будет создан автоматически)
        level: Уровень логирования
        log_args: Логировать аргументы функции
        log_result: Логировать результат функции
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if logger is None:
                func_logger = get_module_logger(func.__module__)
            else:
                func_logger = logger
            
            # Логируем начало выполнения
            if log_args:
                func_logger.log(level, f"Вызов {func.__name__} с аргументами: args={args}, kwargs={kwargs}")
            else:
                func_logger.log(level, f"Вызов {func.__name__}")
            
            try:
                result = func(*args, **kwargs)
                
                if log_result:
                    func_logger.log(level, f"Функция {func.__name__} завершена успешно, результат: {result}")
                else:
                    func_logger.log(level, f"Функция {func.__name__} завершена успешно")
                
                return result
            except Exception as e:
                func_logger.error(f"Error in function {func.__name__}: {e}")
                raise
        
        return wrapper
    return decorator


def log_execution_time(logger: Optional[logging.Logger] = None,
                      level: int = logging.INFO,
                      log_args: bool = False):
    """
    Декоратор для логирования времени выполнения функций
    
    Args:
        logger: Логгер для использования
        level: Уровень логирования
        log_args: Логировать аргументы функции
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if logger is None:
                func_logger = get_module_logger(func.__module__)
            else:
                func_logger = logger
            
            start_time = time.time()
            
            if log_args:
                func_logger.log(level, f"Начало выполнения {func.__name__} с аргументами: args={args}, kwargs={kwargs}")
            else:
                func_logger.log(level, f"Начало выполнения {func.__name__}")
            
            try:
                result = func(*args, **kwargs)
                execution_time = time.time() - start_time
                func_logger.log(level, f"Функция {func.__name__} выполнена за {execution_time:.4f} секунд")
                return result
            except Exception as e:
                execution_time = time.time() - start_time
                func_logger.error(f"Error in function {func.__name__} after {execution_time:.4f} seconds: {e}")
                raise
        
        return wrapper
    return decorator


@contextmanager
def log_context(logger: logging.Logger, 
               message: str, 
               level: int = logging.INFO,
               log_start: bool = True,
               log_end: bool = True):
    """
    Контекстный менеджер для логирования блоков кода
    
    Args:
        logger: Логгер для использования
        message: Сообщение для логирования
        level: Уровень логирования
        log_start: Логировать начало блока
        log_end: Логировать конец блока
    """
    if log_start:
        logger.log(level, f"Начало: {message}")
    
    start_time = time.time()
    
    try:
        yield
    except Exception as e:
        execution_time = time.time() - start_time
        logger.error(f"Error in block '{message}' after {execution_time:.4f} seconds: {e}")
        raise
    finally:
        if log_end:
            execution_time = time.time() - start_time
            logger.log(level, f"Завершение: {message} (время выполнения: {execution_time:.4f} секунд)")


def log_performance_metric(logger: logging.Logger,
                          metric_name: str,
                          value: float,
                          unit: str = "",
                          level: int = logging.INFO):
    """
    Логирует метрику производительности
    
    Args:
        logger: Логгер для использования
        metric_name: Название метрики
        value: Значение метрики
        unit: Единица измерения
        level: Уровень логирования
    """
    unit_str = f" {unit}" if unit else ""
    logger.log(level, f"PERFORMANCE: {metric_name} = {value}{unit_str}")


def log_detection_result(logger: logging.Logger,
                        frame_id: int,
                        source_id: str,
                        detections_count: int,
                        processing_time: float):
    """
    Логирует результат детекции объектов
    
    Args:
        logger: Логгер для использования
        frame_id: ID кадра
        source_id: ID источника
        detections_count: Количество обнаруженных объектов
        processing_time: Время обработки в секундах
    """
    logger.info(f"Detection completed - Frame: {frame_id}, Source: {source_id}, "
                f"Objects: {detections_count}, Time: {processing_time:.4f}s")


def log_tracking_result(logger: logging.Logger,
                       frame_id: int,
                       source_id: str,
                       active_tracks: int,
                       lost_tracks: int,
                       processing_time: float):
    """
    Логирует результат трекинга объектов
    
    Args:
        logger: Логгер для использования
        frame_id: ID кадра
        source_id: ID источника
        active_tracks: Количество активных треков
        lost_tracks: Количество потерянных треков
        processing_time: Время обработки в секундах
    """
    logger.info(f"Tracking completed - Frame: {frame_id}, Source: {source_id}, "
                f"Active tracks: {active_tracks}, Lost: {lost_tracks}, "
                f"Time: {processing_time:.4f}s")


def log_pipeline_step(logger: logging.Logger,
                      step_name: str,
                      input_count: int,
                      output_count: int,
                      processing_time: float,
                      dropped_count: int = 0):
    """
    Логирует выполнение шага пайплайна
    
    Args:
        logger: Логгер для использования
        step_name: Название шага
        input_count: Количество входных элементов
        output_count: Количество выходных элементов
        processing_time: Время обработки в секундах
        dropped_count: Количество отброшенных элементов
    """
    dropped_info = f", Dropped: {dropped_count}" if dropped_count > 0 else ""
    logger.debug(f"Pipeline step '{step_name}': "
                f"Input: {input_count}, Output: {output_count}{dropped_info}, "
                f"Time: {processing_time:.4f}s")


def log_error_with_context(logger: logging.Logger,
                          error: Exception,
                          context: str = "",
                          level: int = logging.ERROR):
    """
    Логирует ошибку с контекстом
    
    Args:
        logger: Логгер для использования
        error: Исключение
        context: Контекст ошибки
        level: Уровень логирования
    """
    context_str = f" в контексте: {context}" if context else ""
    logger.log(level, f"Ошибка{context_str}: {type(error).__name__}: {error}")


def log_system_startup(logger: logging.Logger, config_info: dict):
    """
    Логирует информацию о запуске системы
    
    Args:
        logger: Логгер для использования
        config_info: Информация о конфигурации
    """
    logger.info("=== Starting EvilEye ===")
    logger.info(f"Configuration: {config_info}")
    logger.info("=" * 30)


def log_system_shutdown(logger: logging.Logger, reason: str = "Нормальное завершение"):
    """
    Логирует информацию о завершении системы
    
    Args:
        logger: Логгер для использования
        reason: Причина завершения
    """
    logger.info("=== EvilEye Shutdown ===")
    logger.info(f"Reason: {reason}")
    logger.info("=" * 30)



"""
Visualization adapter for connecting Visualizer to Controller through events.
"""

from typing import Dict, List, Any, Optional
import logging

from .interfaces import IVisualizationProvider
from ..core.logger import get_module_logger


class VisualizationAdapter(IVisualizationProvider):
    """
    Adapter that wraps Visualizer and implements IVisualizationProvider interface.
    
    This adapter allows Controller to use visualization without knowing about GUI details.
    """
    
    def __init__(self, visualizer=None, logger: Optional[logging.Logger] = None):
        """
        Initialize visualization adapter.
        
        Args:
            visualizer: Visualizer instance (can be None if created later)
            logger: Optional logger instance
        """
        self.visualizer = visualizer
        self.logger = logger or get_module_logger("visualization_adapter")
        self._initialized = False
    
    def set_visualizer(self, visualizer) -> None:
        """
        Set visualizer instance.
        
        Args:
            visualizer: Visualizer instance
        """
        self.visualizer = visualizer
        if visualizer:
            self._initialized = True
            self.logger.info("Visualizer set in adapter")
    
    def update(self, processing_frames: Dict[int, Any], 
               source_last_processed_frame_id: Dict[int, int],
               objects: List[Any], 
               dropped_frames: Dict[int, int],
               debug_info: Dict[str, Any]) -> None:
        """
        Update visualization with new frame data.
        
        Args:
            processing_frames: Dictionary mapping source_id to list of CaptureImage frames
                              OR list of CaptureImage frames (will be converted)
            source_last_processed_frame_id: Dictionary mapping source_id to last processed frame ID
            objects: List of ObjectResultList objects
            dropped_frames: Dictionary mapping source_id to dropped frame count
            debug_info: Debug information dictionary
        """
        if self.visualizer:
            try:
                # Convert processing_frames to format expected by visualizer.update()
                # visualizer.update() expects list[CaptureImage], not dict
                processing_frames_list = []
                if isinstance(processing_frames, dict):
                    # Convert dict to flat list
                    for source_id, frames in processing_frames.items():
                        if isinstance(frames, list):
                            processing_frames_list.extend(frames)
                        else:
                            processing_frames_list.append(frames)
                elif isinstance(processing_frames, list):
                    # Already a list
                    processing_frames_list = processing_frames
                
                # Convert dropped_frames to list format expected by visualizer
                # visualizer.update() expects list, but we might have dict[int, int]
                dropped_frames_list = dropped_frames
                if isinstance(dropped_frames, dict):
                    # Convert dict to list - visualizer expects list format
                    # For now, pass empty list if dict format (visualizer might not use it)
                    dropped_frames_list = []
                
                self.visualizer.update(
                    processing_frames_list,
                    source_last_processed_frame_id,
                    objects,
                    dropped_frames_list,
                    debug_info
                )
            except Exception as e:
                self.logger.error(f"Error updating visualizer: {e}", exc_info=True)
    
    def start(self) -> None:
        """Start visualization."""
        if self.visualizer:
            try:
                self.visualizer.start()
                self.logger.info("Visualizer started via adapter")
            except Exception as e:
                self.logger.error(f"Error starting visualizer: {e}", exc_info=True)
    
    def stop(self) -> None:
        """Stop visualization."""
        if self.visualizer:
            try:
                self.visualizer.stop()
                self.logger.info("Visualizer stopped via adapter")
            except Exception as e:
                self.logger.error(f"Error stopping visualizer: {e}", exc_info=True)
    
    def set_params(self, **params) -> None:
        """Set visualization parameters."""
        if self.visualizer:
            try:
                self.visualizer.set_params(**params)
            except Exception as e:
                self.logger.error(f"Error setting visualizer params: {e}", exc_info=True)
    
    def get_params(self) -> Dict[str, Any]:
        """Get current visualization parameters."""
        if self.visualizer:
            try:
                return self.visualizer.get_params()
            except Exception as e:
                self.logger.error(f"Error getting visualizer params: {e}", exc_info=True)
        return {}
    
    def is_available(self) -> bool:
        """Check if visualizer is available."""
        return self.visualizer is not None and self._initialized
    
    def set_current_main_widget_size(self, width: int, height: int) -> None:
        """Set main widget size for visualizer."""
        if self.visualizer and hasattr(self.visualizer, 'set_current_main_widget_size'):
            try:
                self.visualizer.set_current_main_widget_size(width, height)
            except Exception as e:
                self.logger.error(f"Error setting widget size: {e}", exc_info=True)



#!/usr/bin/env python3
"""
Wrapper for EvilEye API server that ensures proper path setup and stable entry point
"""
import sys
import os
from pathlib import Path

# Add project root to Python path
project_root = Path(__file__).parent.parent
if project_root.exists():
    sys.path.insert(0, str(project_root))

from evileye.server import main


def main_wrapper() -> int:
    main()
    return 0


def run() -> None:
    sys.exit(main_wrapper())


def start() -> None:
    sys.exit(main_wrapper())


def cli() -> None:
    sys.exit(main_wrapper())


if __name__ == '__main__':
    sys.exit(main_wrapper())



"""
Unit tests for video seeking functionality in stream player
"""
import unittest
from unittest.mock import Mock, MagicMock, patch, call
import datetime
from pathlib import Path
import sys

# Add parent directory to path
sys.path.insert(0, str(Path(__file__).parent.parent))

try:
    from PyQt6.QtWidgets import QApplication
    from PyQt6.QtCore import Qt
    pyqt_version = 6
except ImportError:
    from PyQt5.QtWidgets import QApplication
    from PyQt5.QtCore import Qt
    pyqt_version = 5

from evileye.visualization_modules.stream_player_components import VideoGridWidget, SplitVideoPlayerWidget
from evileye.visualization_modules.video_player_window import VideoPlayerWidget


class TestStreamPlayerSeeking(unittest.TestCase):
    """Tests for video seeking functionality"""
    
    @classmethod
    def setUpClass(cls):
        """Create QApplication once for all tests"""
        if not QApplication.instance():
            cls.app = QApplication([])
        else:
            cls.app = QApplication.instance()
    
    def setUp(self):
        """Set up test fixtures"""
        self.video_grid = VideoGridWidget()
        
        # Mock start time
        self.start_time = datetime.datetime(2026, 1, 8, 14, 20, 16)
        self.video_grid._start_time = self.start_time
        
        # Create mock players
        self.mock_player1 = Mock(spec=VideoPlayerWidget)
        self.mock_player1.video_path = "/path/to/cam1/video.mp4"
        self.mock_player1._is_playing = True
        self.mock_player1.player = None
        self.mock_player1.cap = Mock()
        self.mock_player1.cap.get.return_value = 30.0  # FPS
        self.mock_player1.timer = Mock()
        self.mock_player1.timer.isActive.return_value = True
        
        self.mock_split_player = Mock(spec=SplitVideoPlayerWidget)
        self.mock_split_player._video_player = Mock()
        self.mock_split_player._video_player.video_path = "/path/to/cam2-cam3/video.mp4"
        self.mock_split_player._video_player._is_playing = True
        self.mock_split_player._video_player.player = None
        self.mock_split_player._video_player.cap = Mock()
        self.mock_split_player._video_player.cap.get.return_value = 30.0
        self.mock_split_player._video_player.timer = Mock()
        self.mock_split_player._video_player.timer.isActive.return_value = True
        self.mock_split_player._region_widgets = []
        
        # Set up video players dictionary
        self.video_grid._video_players = {
            "Cam1": self.mock_player1,
            "Cam2-Cam3": self.mock_split_player
        }
        
        # Set up camera segments with different start times
        self.video_grid._camera_segments = {
            "Cam1": [
                (datetime.datetime(2026, 1, 8, 14, 20, 16), 
                 datetime.datetime(2026, 1, 8, 14, 25, 16), 
                 "/path/to/cam1/video.mp4")
            ],
            "Cam2-Cam3": [
                (datetime.datetime(2026, 1, 8, 14, 20, 16), 
                 datetime.datetime(2026, 1, 8, 14, 25, 16), 
                 "/path/to/cam2-cam3/video.mp4")
            ],
            "Cam4-Cam5": [
                (datetime.datetime(2026, 1, 8, 14, 20, 19),  # Started 3 seconds later
                 datetime.datetime(2026, 1, 8, 14, 25, 19), 
                 "/path/to/cam4-cam5/video.mp4")
            ]
        }
        
        # Set up folder to sources mapping
        self.video_grid._folder_to_sources = {
            "Cam1": ["Cam1"],
            "Cam2-Cam3": ["Cam2", "Cam3"],
            "Cam4-Cam5": ["Cam4", "Cam5"]
        }
        
        # Set up current segments and indices
        self.video_grid._current_segments = {
            "Cam1": "/path/to/cam1/video.mp4",
            "Cam2-Cam3": "/path/to/cam2-cam3/video.mp4"
        }
        self.video_grid._current_segment_indices = {
            "Cam1": 0,
            "Cam2-Cam3": 0
        }
        
        # Set up source config
        self.video_grid._source_config = {
            "Cam2-Cam3": {
                "split": True,
                "num_split": 2,
                "src_coords": [[0, 0, 960, 540], [960, 0, 960, 540]],
                "source_names": ["Cam2", "Cam3"],
                "parent_folder": "Cam2-Cam3"
            },
            "Cam4-Cam5": {
                "split": True,
                "num_split": 2,
                "src_coords": [[0, 0, 960, 540], [960, 0, 960, 540]],
                "source_names": ["Cam4", "Cam5"],
                "parent_folder": "Cam4-Cam5"
            }
        }
    
    def test_seek_all_calls_seek_for_all_cameras(self):
        """Test that _seek_player is called for all cameras"""
        # Add Cam4-Cam5 player
        mock_player4 = Mock(spec=SplitVideoPlayerWidget)
        mock_player4._video_player = Mock()
        mock_player4._video_player.video_path = "/path/to/cam4-cam5/video.mp4"
        mock_player4._video_player._is_playing = True
        mock_player4._video_player.player = None
        mock_player4._video_player.cap = Mock()
        mock_player4._video_player.cap.get.return_value = 30.0
        mock_player4._video_player.timer = Mock()
        mock_player4._video_player.timer.isActive.return_value = True
        mock_player4._region_widgets = []
        self.video_grid._video_players["Cam4-Cam5"] = mock_player4
        self.video_grid._current_segments["Cam4-Cam5"] = "/path/to/cam4-cam5/video.mp4"
        self.video_grid._current_segment_indices["Cam4-Cam5"] = 0
        
        # Mock _seek_player to track calls
        seek_calls = []
        original_seek = self.video_grid._seek_player
        
        def track_seek(player, position_ms):
            seek_calls.append((type(player).__name__, position_ms))
            return original_seek(player, position_ms)
        
        self.video_grid._seek_player = track_seek
        
        # Call seek_all
        self.video_grid.seek_all(1000, should_play=True)
        
        # Check that _seek_player was called for all cameras
        self.assertEqual(len(seek_calls), 3, f"Expected 3 seek calls, got {len(seek_calls)}: {seek_calls}")
        
        # Verify calls for each camera
        camera_names = [call[0] for call in seek_calls]
        self.assertIn("VideoPlayerWidget", camera_names, "Cam1 should be seeked")
        self.assertIn("SplitVideoPlayerWidget", camera_names, "Split players should be seeked")
    
    def test_seek_all_calculates_correct_offset(self):
        """Test that segment_offset_ms is calculated correctly for each camera"""
        # Add Cam4-Cam5 player
        mock_player4 = Mock(spec=SplitVideoPlayerWidget)
        mock_player4._video_player = Mock()
        mock_player4._video_player.video_path = "/path/to/cam4-cam5/video.mp4"
        mock_player4._video_player._is_playing = True
        mock_player4._video_player.player = None
        mock_player4._video_player.cap = Mock()
        mock_player4._video_player.cap.get.return_value = 30.0
        mock_player4._video_player.timer = Mock()
        mock_player4._video_player.timer.isActive.return_value = True
        mock_player4._region_widgets = []
        self.video_grid._video_players["Cam4-Cam5"] = mock_player4
        self.video_grid._current_segments["Cam4-Cam5"] = "/path/to/cam4-cam5/video.mp4"
        self.video_grid._current_segment_indices["Cam4-Cam5"] = 0
        
        # Track positions passed to _seek_player
        seek_positions = {}
        
        def track_seek(player, position_ms):
            camera_name = None
            for name, p in self.video_grid._video_players.items():
                if p == player:
                    camera_name = name
                    break
            if camera_name:
                seek_positions[camera_name] = position_ms
        
        self.video_grid._seek_player = track_seek
        
        # Seek to 2000ms (2 seconds after start)
        self.video_grid.seek_all(2000, should_play=True)
        
        # Cam1 started at 14:20:16, so at 2000ms offset should be 2000ms
        self.assertEqual(seek_positions.get("Cam1"), 2000, 
                        f"Cam1 should seek to 2000ms, got {seek_positions.get('Cam1')}")
        
        # Cam4-Cam5 started at 14:20:19 (3 seconds later), so at 2000ms global time
        # target_time = 14:20:18, segment_start = 14:20:19
        # segment_offset = -1000ms, but should be clamped to 0ms
        self.assertEqual(seek_positions.get("Cam4-Cam5"), 0,
                        f"Cam4-Cam5 should seek to 0ms (started later), got {seek_positions.get('Cam4-Cam5')}")
    
    def test_seek_all_handles_late_starting_cameras(self):
        """Test handling of cameras that started recording later"""
        # Add Cam4-Cam5 player
        mock_player4 = Mock(spec=SplitVideoPlayerWidget)
        mock_player4._video_player = Mock()
        mock_player4._video_player.video_path = "/path/to/cam4-cam5/video.mp4"
        mock_player4._video_player._is_playing = True
        mock_player4._video_player.player = None
        mock_player4._video_player.cap = Mock()
        mock_player4._video_player.cap.get.return_value = 30.0
        mock_player4._video_player.timer = Mock()
        mock_player4._video_player.timer.isActive.return_value = True
        mock_player4._region_widgets = []
        self.video_grid._video_players["Cam4-Cam5"] = mock_player4
        self.video_grid._current_segments["Cam4-Cam5"] = "/path/to/cam4-cam5/video.mp4"
        self.video_grid._current_segment_indices["Cam4-Cam5"] = 0
        
        # Seek to 5000ms (5 seconds after start)
        # Cam4-Cam5 started at 14:20:19, so target_time (14:20:21) is after segment_start (14:20:19)
        # segment_offset should be 2000ms (5s - 3s delay)
        
        seek_positions = {}
        
        def track_seek(player, position_ms):
            camera_name = None
            for name, p in self.video_grid._video_players.items():
                if p == player:
                    camera_name = name
                    break
            if camera_name:
                seek_positions[camera_name] = position_ms
        
        self.video_grid._seek_player = track_seek
        
        self.video_grid.seek_all(5000, should_play=True)
        
        # Cam4-Cam5 should seek to 2000ms (5000ms - 3000ms delay)
        self.assertEqual(seek_positions.get("Cam4-Cam5"), 2000,
                        f"Cam4-Cam5 should seek to 2000ms (5s - 3s delay), got {seek_positions.get('Cam4-Cam5')}")
    
    def test_seek_player_for_regular_player(self):
        """Test _seek_player for regular VideoPlayerWidget"""
        # Test with OpenCV backend
        position_ms = 1500
        self.video_grid._seek_player(self.mock_player1, position_ms)
        
        # Verify cap.set was called with correct frame number
        expected_frame = int((position_ms / 1000.0) * 30)
        self.mock_player1.cap.set.assert_called_once()
        call_args = self.mock_player1.cap.set.call_args
        self.assertEqual(call_args[0][0], 0)  # CAP_PROP_POS_FRAMES
        self.assertEqual(call_args[0][1], expected_frame)
    
    def test_seek_player_for_split_player(self):
        """Test _seek_player for SplitVideoPlayerWidget"""
        position_ms = 2000
        self.mock_split_player.seek = Mock()
        
        self.video_grid._seek_player(self.mock_split_player, position_ms)
        
        # Verify seek was called on split player
        self.mock_split_player.seek.assert_called_once_with(position_ms)
    
    def test_seek_all_synchronizes_all_cameras(self):
        """Test that seek_all synchronizes all cameras"""
        # Add Cam4-Cam5 player
        mock_player4 = Mock(spec=SplitVideoPlayerWidget)
        mock_player4._video_player = Mock()
        mock_player4._video_player.video_path = "/path/to/cam4-cam5/video.mp4"
        mock_player4._video_player._is_playing = True
        mock_player4._video_player.player = None
        mock_player4._video_player.cap = Mock()
        mock_player4._video_player.cap.get.return_value = 30.0
        mock_player4._video_player.timer = Mock()
        mock_player4._video_player.timer.isActive.return_value = True
        mock_player4._region_widgets = []
        self.video_grid._video_players["Cam4-Cam5"] = mock_player4
        self.video_grid._current_segments["Cam4-Cam5"] = "/path/to/cam4-cam5/video.mp4"
        self.video_grid._current_segment_indices["Cam4-Cam5"] = 0
        
        # Track all seek calls
        seek_calls = []
        
        def track_seek(player, position_ms):
            camera_name = None
            for name, p in self.video_grid._video_players.items():
                if p == player:
                    camera_name = name
                    break
            seek_calls.append((camera_name, position_ms))
        
        self.video_grid._seek_player = track_seek
        
        # Seek to 3000ms
        self.video_grid.seek_all(3000, should_play=True)
        
        # All cameras should have been seeked
        self.assertEqual(len(seek_calls), 3, 
                        f"Expected 3 cameras to be seeked, got {len(seek_calls)}")
        
        # Verify all cameras are in the calls
        camera_names = [call[0] for call in seek_calls]
        self.assertIn("Cam1", camera_names)
        self.assertIn("Cam2-Cam3", camera_names)
        self.assertIn("Cam4-Cam5", camera_names)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import datetime
from typing import List
from ultralytics.engine.results import Boxes
from ultralytics.trackers.bot_sort import BOTrack
from .object_tracking_base import ObjectTrackingBase
from .trackers.bot_sort import BOTSORT
from .trackers.track_encoder import TrackEncoder
from .trackers.cfg.utils import read_cfg
from time import sleep
from ..object_detector.object_detection_base import DetectionResult
from ..object_detector.object_detection_base import DetectionResultList
from .tracking_results import TrackingResult
from .tracking_results import TrackingResultList
from dataclasses import dataclass
import copy
from ..core.base_class import EvilEyeBase

@dataclass
class BostSortCfg:
    appearance_thresh: float = 0.25
    gmc_method: str = "sparseOptFlow"
    match_thresh: float = 0.8
    new_track_thresh: float = 0.6
    proximity_thresh: float = 0.5
    track_buffer: int = 30
    track_high_thresh: float = 0.5
    track_low_thresh: float = 0.1
    tracker_type: str = "botsort"
    fuse_score: bool = True
    with_reid: bool = False



@EvilEyeBase.register("ObjectTrackingBotsort")
class ObjectTrackingBotsort(ObjectTrackingBase):
    #tracker: BOTSORT

    def __init__(self):
        super().__init__()
        self.botsort_cfg = BostSortCfg()
        self.tracker = None
        self.encoders = None
        self.fps = 5

        self.cfg_dict = dict()
        self.cfg_dict["appearance_thresh"] = 0.25
        self.cfg_dict["gmc_method"] = "sparseOptFlow"
        self.cfg_dict["match_thresh"] = 0.8
        self.cfg_dict["new_track_thresh"] = 0.6
        self.cfg_dict["proximity_thresh"] = 0.5
        self.cfg_dict["track_buffer"] = 30
        self.cfg_dict["track_high_thresh"] = 0.5
        self.cfg_dict["track_low_thresh"] = 0.1
        self.cfg_dict["tracker_type"] = "botsort"
        self.cfg_dict["with_reid"] = False

    def init_impl(self, **kwargs):
        try:
            encoders = kwargs.get('encoders', None)
            if encoders is not None:
                onnx_path = self.params.get("tracker_onnx", "models/osnet_ain_x1_0_M.onnx")
                if onnx_path in encoders:
                    encoder = encoders[onnx_path]
                    self.encoders = [encoder]
                    self.logger.debug(f"Using encoder from encoders dict: {onnx_path}")
                else:
                    self.encoders = None
                    self.logger.debug(f"Encoder {onnx_path} not found in encoders dict, ReID disabled")
            else:
                self.encoders = None
                self.logger.debug("No encoders provided, ReID disabled")
            
            super().init_impl(**kwargs)
            
            # Ensure botsort_cfg is set (should be set by set_params_impl, but check anyway)
            if not self.botsort_cfg:
                # Try to set default config if not set
                self.logger.warning("botsort_cfg not set, using default configuration")
                self.botsort_cfg = BostSortCfg()
            
            self.logger.debug(f"Initializing BOTSORT with fps={self.fps}, with_reid={self.botsort_cfg.with_reid}")
            self.tracker = BOTSORT(self.botsort_cfg, self.encoders, frame_rate=self.fps)
            self.logger.debug("BOTSORT tracker initialized successfully")
            return True
        except Exception as e:
            self.logger.error(f"Failed to initialize ObjectTrackingBotsort: {e}", exc_info=True)
            self.tracker = None
            return False

    def release_impl(self):
        super().init_impl()
        self.tracker = None

    def reset_impl(self):
        self.tracker.reset()

    def set_params_impl(self):
        self.source_ids = self.params.get('source_ids', [])
        self.fps = self.params.get('fps', 5)

        self.cfg_dict = self.params.get('botsort_cfg', self.cfg_dict)
        cfg_dict = self.cfg_dict

        if cfg_dict:
            self.botsort_cfg = BostSortCfg(appearance_thresh=cfg_dict["appearance_thresh"], gmc_method=cfg_dict["gmc_method"],
                                           match_thresh=cfg_dict["match_thresh"], new_track_thresh=cfg_dict["new_track_thresh"],
                                           proximity_thresh=cfg_dict["proximity_thresh"], track_buffer=cfg_dict["track_buffer"],
                                           track_high_thresh=cfg_dict["track_high_thresh"], track_low_thresh=cfg_dict["track_low_thresh"],
                                           tracker_type=cfg_dict["tracker_type"], with_reid=cfg_dict["with_reid"])

    def get_params_impl(self):
        params = dict()
        params['source_ids'] = self.source_ids
        params['fps'] = self.fps
        params['botsort_cfg'] = self.cfg_dict
        return params

    def default(self):
        self.params.clear()

    def _process_impl(self):
        while self.run_flag:
            sleep(0.01)
            detections = self.queue_in.get()
            if detections is None:
                continue
            if self.tracker is None:
                continue
            detection_result, image = detections
            
            # Check if image is valid
            if image is None or image.image is None:
                self.logger.warning(f"Received None image for source {detection_result.source_id if detection_result else 'unknown'}, skipping")
                continue
            
            try:
                cam_id, boxes = self._parse_det_info(detection_result, image.image)
                tracks = self.tracker.update(boxes, image.image)
                if len(tracks) > 0:
                    pass
                tracks_info = self._create_tracks_info(cam_id, detection_result.frame_id, None, tracks)
                self.queue_out.put((tracks_info, image))
            except Exception as e:
                self.logger.error(f"Error processing detection for source {detection_result.source_id if detection_result else 'unknown'}: {e}", exc_info=True)
                continue

    def _parse_det_info(self, det_info: DetectionResultList, image: np.ndarray) -> tuple:
        if image is None:
            raise ValueError("image cannot be None")
        
        cam_id = det_info.source_id
        objects = det_info.detections

        bboxes_xyxy = []
        confidences = []
        class_ids = []

        for obj in objects:
            bboxes_xyxy.append(obj.bounding_box)
            confidences.append(obj.confidence)
            class_ids.append(obj.class_id)

        bboxes_xyxy = np.array(bboxes_xyxy).reshape(-1, 4)
        confidences = np.array(confidences).reshape(-1, 1)
        class_ids = np.array(class_ids).reshape(-1, 1)

        bboxes_xyxy = np.array(bboxes_xyxy)
        confidences = np.array(confidences)
        class_ids = np.array(class_ids)
        
        boxes_array = np.concatenate([bboxes_xyxy, confidences, class_ids], axis=1)
        
        # Validate image shape
        if not hasattr(image, 'shape') or len(image.shape) < 2:
            raise ValueError(f"Invalid image shape: {image.shape if hasattr(image, 'shape') else 'no shape attribute'}")
        
        orig_shape = (image.shape[1], image.shape[0])
        boxes = Boxes(boxes_array, orig_shape)
        return cam_id, boxes

    def _create_tracks_info(
            self, 
            cam_id: int, 
            frame_id: int, 
            detection: DetectionResult, 
            tracks: list[BOTrack]):
        
        tracks_info = TrackingResultList()
        tracks_info.source_id = cam_id
        tracks_info.frame_id = frame_id
        tracks_info.time_stamp = datetime.datetime.now()

        # print(tracks)
        tracks_results = np.asarray([x.result for x in tracks], dtype=np.float32)
        for i in range(len(tracks_results)):
            track_bbox = tracks_results[i, :4].tolist()
            track_conf = tracks_results[i, 5]
            track_cls = int(tracks_results[i, 6])
            track_id = int(tracks_results[i, 4])
            object_info = TrackingResult()
            object_info.class_id = track_cls
            object_info.bounding_box = track_bbox
            object_info.confidence = float(track_conf)
            object_info.track_id = track_id
            if detection:
                object_info.detection_history.append(detection)
            
            # Add BOTrack object to tracking data
            # in order to use it in multi-camera tracking during reidentification
            object_info.tracking_data = {
                "track_object": tracks[i],
            }

            tracks_info.tracks.append(object_info)

        return tracks_info




import time

from .database_controller_pg import DatabaseControllerPg
from .db_adapter import DatabaseAdapterBase
import json
from ..utils.utils import ObjectResultEncoder
import copy
import datetime
import os
from timeit import default_timer as timer
import cv2
from ..utils import threading_events
from ..utils import utils
from psycopg2 import sql


class DatabaseAdapterObjects(DatabaseAdapterBase):
    def __init__(self, db_controller):
        super().__init__(db_controller)
        self.image_dir = self.db_params['image_dir']
        self.preview_width = self.db_params['preview_width']
        self.preview_height = self.db_params['preview_height']
        self.preview_size = (self.preview_width, self.preview_height)

    def _insert_impl(self, obj):
        fields, data, preview_path, frame_path = self._prepare_for_saving(obj)
        query_type = 'insert'
        insert_query = sql.SQL("INSERT INTO {} ({}) VALUES ({}) RETURNING record_id, bounding_box").format(
            sql.Identifier('objects'),
            sql.SQL(",").join(map(sql.Identifier, fields)),
            sql.SQL(', ').join(sql.Placeholder() * len(fields))
        )
        self.queue_in.put((query_type, insert_query, data, preview_path, frame_path, obj.last_image))

    def _update_impl(self, obj):
        fields, data, preview_path, frame_path = self._prepare_for_updating(obj)

        query_type = 'Update'
        data = list(data)
        data.append(obj.object_id)
        data = tuple(data)
        table_name = 'objects'
        last_obj_query = sql.SQL('''SELECT distinct on (object_id) record_id FROM {table} WHERE object_id = {id} 
                                    ORDER BY object_id, record_id DESC''').format(
            id=sql.Placeholder(),
            fields=sql.SQL(",").join(map(sql.Identifier, fields)),
            table=sql.Identifier(table_name))
        update_query = sql.SQL('UPDATE {table} SET {data} WHERE record_id=({selected}) '
                               'RETURNING record_id, lost_bounding_box').format(
            table=sql.Identifier(table_name),
            data=sql.SQL(', ').join(
                sql.Composed([sql.Identifier(field), sql.SQL(" = "), sql.Placeholder()]) for field in fields),
            selected=sql.Composed(last_obj_query),
            fields=sql.SQL(",").join(map(sql.Identifier, fields)))
        self.queue_in.put((query_type, update_query, data, preview_path, frame_path, obj.last_image))

    def _execute_query(self):
        while self.run_flag:
            time.sleep(0.01)
            try:
                if not self.queue_in.empty():
                    query_type, query_string, data, preview_path, frame_path, image = self.queue_in.get()
                    if query_string is not None:
                        pass
                else:
                    query_type = query_string = data = preview_path = frame_path = image = None
            except ValueError:
                break

            if query_string is None:
                continue

            # Если контроллер БД не подключен, аккуратно пропускаем запись
            if (
                not hasattr(self.db_controller, "is_connected")
                or not self.db_controller.is_connected()
            ):
                self.logger.warning(
                    "Database is not connected in db_adapter_objects._execute_query; "
                    "skipping DB write operation."
                )
                continue

            record = self.db_controller.query(query_string, data)

            # query() мог вернуть None (ошибка БД или нет данных) — проверяем это
            if not record:
                self.logger.warning(
                    "Database query returned no records in db_adapter_objects._execute_query; "
                    "skipping image save and notifications."
                )
                continue

            try:
                row_num = record[0][0]
                box = record[0][1]
            except Exception as ex:
                self.logger.error(
                    f"Unexpected record format in db_adapter_objects._execute_query: {record}. "
                    f"Error: {ex}"
                )
                continue

            start_save_it = timer()
            self._save_image(preview_path, frame_path, image, box)
            end_save_it = timer()
            if query_type == 'insert':
                threading_events.notify('handler new object')
            elif query_type == 'update':
                threading_events.notify('handler update object')

    def _save_image(self, preview_path, frame_path, image, box):
        preview_save_dir = os.path.join(self.image_dir, preview_path)
        frame_save_dir = os.path.join(self.image_dir, frame_path)
        # Save clean preview without overlays
        preview = cv2.resize(copy.deepcopy(image.image), self.preview_size, cv2.INTER_NEAREST)
        preview_saved = cv2.imwrite(preview_save_dir, preview)
        # Save original frame without overlays
        frame_saved = cv2.imwrite(frame_save_dir, image.image)
        if not preview_saved or not frame_saved:
            self.logger.error(f'ERROR: can\'t save image file {frame_save_dir}')

    def _prepare_for_updating(self, obj):
        fields_for_updating = {'lost_bounding_box': obj.track.bounding_box,
                               'time_lost': obj.time_lost,
                               'lost_preview_path': '',
                               'lost_frame_path': '',
                               'object_data': json.dumps(obj.__dict__, cls=ObjectResultEncoder)}

        src_name = ''
        for camera in self.cameras_params:
            if obj.source_id in camera['source_ids']:
                id_idx = camera['source_ids'].index(obj.source_id)
                src_name = camera['source_names'][id_idx]
                break
        fields_for_updating['lost_preview_path'] = self._get_img_path('preview', 'lost',
                                                                      src_name, obj)
        fields_for_updating['lost_frame_path'] = self._get_img_path('frame', 'lost',
                                                                    src_name, obj)

        image_height, image_width, _ = obj.last_image.image.shape
        fields_for_updating['lost_bounding_box'] = copy.deepcopy(fields_for_updating['lost_bounding_box'])
        fields_for_updating['lost_bounding_box'][0] /= image_width
        fields_for_updating['lost_bounding_box'][1] /= image_height
        fields_for_updating['lost_bounding_box'][2] /= image_width
        fields_for_updating['lost_bounding_box'][3] /= image_height
        return (list(fields_for_updating.keys()), list(fields_for_updating.values()),
                fields_for_updating['lost_preview_path'], fields_for_updating['lost_frame_path'])

    def _prepare_for_saving(self, obj) -> tuple[list, list, str, str]:
        fields_for_saving = {'source_id': obj.source_id,
                             'source_name': '',
                             'time_stamp': obj.time_detected,
                             'time_lost': obj.time_lost,
                             'object_id': obj.object_id,
                             'bounding_box': obj.track.bounding_box,
                             'lost_bounding_box': None,
                             'confidence': obj.track.confidence,
                             'class_id': obj.class_id,
                             'preview_path': '',
                             'lost_preview_path': None,
                             'frame_path': '',
                             'lost_frame_path': None,
                             'object_data': json.dumps(obj.__dict__, cls=ObjectResultEncoder),
                             'project_id': self.db_controller.get_project_id(),
                             'job_id': self.db_controller.get_job_id(),
                             'camera_full_address': ''}

        for camera in self.cameras_params:
            if obj.source_id in camera['source_ids']:
                id_idx = camera['source_ids'].index(obj.source_id)
                fields_for_saving['source_name'] = camera['source_names'][id_idx]
                fields_for_saving['camera_full_address'] = camera['camera']
                break
        fields_for_saving['preview_path'] = self._get_img_path('preview', 'detected',
                                                               fields_for_saving['source_name'], obj)
        fields_for_saving['frame_path'] = self._get_img_path('frame', 'detected',
                                                             fields_for_saving['source_name'], obj)

        image_height, image_width, _ = obj.last_image.image.shape
        fields_for_saving['bounding_box'] = copy.deepcopy(fields_for_saving['bounding_box'])
        fields_for_saving['bounding_box'][0] /= image_width
        fields_for_saving['bounding_box'][1] /= image_height
        fields_for_saving['bounding_box'][2] /= image_width
        fields_for_saving['bounding_box'][3] /= image_height
        return (list(fields_for_saving.keys()), list(fields_for_saving.values()),
                fields_for_saving['preview_path'], fields_for_saving['frame_path'])

    def _get_img_path(self, image_type, obj_event_type, src_name, obj):
        save_dir = self.db_params['image_dir']
        detections_dir = os.path.join(save_dir, 'Detections')
        cur_date = datetime.date.today()
        cur_date_str = cur_date.strftime('%Y-%m-%d')

        current_day_path = os.path.join(detections_dir, cur_date_str)
        images_dir = os.path.join(current_day_path, 'Images')
        # New folders for objects: FoundFrames/FoundPreviews/LostFrames/LostPreviews
        if obj_event_type == 'detected':
            if image_type == 'preview':
                subdir = 'FoundPreviews'
            else:
                subdir = 'FoundFrames'
        else:  # lost
            if image_type == 'preview':
                subdir = 'LostPreviews'
            else:
                subdir = 'LostFrames'
        obj_type_path = os.path.join(images_dir, subdir)
        # obj_event_path = os.path.join(current_day_path, obj_event_type)
        os.makedirs(detections_dir, exist_ok=True)
        os.makedirs(current_day_path, exist_ok=True)
        os.makedirs(images_dir, exist_ok=True)
        os.makedirs(obj_type_path, exist_ok=True)
        # if not os.path.exists(obj_event_path):
        #     os.mkdir(obj_event_path)

        if obj_event_type == 'detected':
            timestamp = obj.time_detected.strftime('%Y-%m-%d_%H-%M-%S.%f')
            img_path = os.path.join(obj_type_path, f'{timestamp}_{src_name}_{image_type}.jpeg')
        elif obj_event_type == 'lost':
            timestamp = obj.time_lost.strftime('%Y-%m-%d_%H-%M-%S-%f')
            img_path = os.path.join(obj_type_path, f'{timestamp}_{src_name}_{image_type}.jpeg')
        return os.path.relpath(img_path, save_dir)



from abc import ABC, abstractmethod
from ..core.base_class import EvilEyeBase
from queue import Queue
import threading
from .tracking_results import TrackingResultList


class ObjectTrackingBase(EvilEyeBase):
    ResultType = TrackingResultList

    def __init__(self):
        super().__init__()

        self.run_flag = False
        self.queue_in = Queue(maxsize=2)
        self.queue_out = Queue()
        self.queue_dropped_id = Queue()
        self.source_ids = []
        self.processing_thread = None

    def put(self, det_info, force=False):
        dropped_id = []
        result = True
        if self.queue_in.full():
            if force:
                dropped_data = self.queue_in.get()
                dropped_id.append(dropped_data[1].source_id)
                dropped_id.append(dropped_data[1].frame_id)
                result = True
            else:
                dropped_id.append(det_info[1].source_id)
                dropped_id.append(det_info[1].frame_id)
                result = False
        if len(dropped_id) > 0:
            self.queue_dropped_id.put(dropped_id)

        if result:
            self.queue_in.put(det_info)

        return result

    def get(self):
        if self.queue_out.empty():
            return None
        return self.queue_out.get()

    def get_dropped_ids(self) -> list:
        res = []
        while not self.queue_dropped_id.empty():
            res.append(self.queue_dropped_id.get())
        return res

    def get_oueue_out_size(self):
        return self.queue_out.qsize()

    def get_source_ids(self):
        return self.source_ids

    def start(self):
        self.run_flag = True
        self.processing_thread.start()

    def stop(self):
        self.run_flag = False
        self.queue_in.put(None)
        if self.processing_thread and self.processing_thread.is_alive():
            self.processing_thread.join()
        self.logger.info('Tracker stopped')

    def init_impl(self, **kwargs):
        self.processing_thread = threading.Thread(target=self._process_impl)

    def release_impl(self):
        del self.processing_thread
        self.processing_thread = None

    @abstractmethod
    def _process_impl(self):
        pass



import threading
import os
import subprocess
import multiprocessing
from typing import Optional, Dict
from evileye.core.logger import get_module_logger
from evileye.controller.controller import Controller
from evileye.api.core.broker_access import get_broker

""" Module for managing pipelines """

class PipelineState:
    CREATED = "created"
    STARTING = "starting"
    RUNNING = "running"
    STOPPING = "stopping"
    STOPPED = "stopped"
    ERROR = "error"


class PipelineRunner:
    def __init__(self, pipeline_id: int, config_body: dict, name: str = None):
        self.pipeline_id = pipeline_id
        self.config_body = config_body
        self.pipeline_name = name or f"Pipeline-{pipeline_id}" 
        self.controller = None
        self.state = PipelineState.CREATED
        self.error: Optional[str] = None
        self.logger = get_module_logger("api.pipeline_runner")

    def start(self) -> None:
        if self.state in (PipelineState.RUNNING, PipelineState.STARTING):
            self.logger.info(f"Pipeline '{self.pipeline_id}' already running or starting")
            return
        try:
            os.environ["EVILEYE_PIPELINE_ID"] = str(self.pipeline_id)
            self.logger.info(f"Set EVILEYE_PIPELINE_ID={self.pipeline_id} for controller")            
            # Ensure no GUI in API mode
            controller_cfg = self.config_body.setdefault("controller", {})
            controller_cfg["gui_enabled"] = False
            controller_cfg["show_main_gui"] = False
            pipeline_cfg = self.config_body.setdefault("pipeline", {})
            pipeline_cfg["name"] = self.pipeline_name
            self.logger.info(f"Set pipeline name to '{self.pipeline_name}' for streaming")

            if self.controller is None:
                self.controller = Controller()
                self.controller.init(self.config_body)

            self.state = PipelineState.STARTING
            self.logger.info(f"Starting controller for pipeline '{self.pipeline_id}'")
            self.controller.start()
            self.logger.info(f"Controller.start() completed for pipeline '{self.pipeline_id}'")
            # Give controller a moment to initialize
            import time
            time.sleep(5.0)  
            self.state = PipelineState.RUNNING
            self.logger.info(f"Pipeline '{self.pipeline_id}' running")

        except Exception as e:
            self.state = PipelineState.ERROR
            self.error = str(e)
            self.logger.error(f"Pipeline '{self.pipeline_id}' failed to start: {e}")


    def stop(self) -> None:
        self.logger.info(f"Pipeline '{self.pipeline_id}' stop() called")
        if self.controller is None:
            self.state = PipelineState.STOPPED
            self.logger.info(f"Pipeline '{self.pipeline_id}' stopped (no controller)")
            return
        if self.state in (PipelineState.STOPPING, PipelineState.STOPPED):
            self.logger.info(f"Pipeline '{self.pipeline_id}' already stopping or stopped")
            return
        self.state = PipelineState.STOPPING
        self.logger.info(f"Pipeline '{self.pipeline_id}' state set to STOPPING")
        try:
            if self.controller:
                self.logger.info(f"Stopping controller for pipeline '{self.pipeline_id}'")
                try:
                    self.controller.stop()
                    self.logger.info(f"Controller stopped for pipeline '{self.pipeline_id}'")
                except Exception as e:
                    self.logger.error(f"Error stopping controller: {e}")
                
                self.logger.info(f"Releasing controller for pipeline '{self.pipeline_id}'")
                try:
                    self.controller.release()
                    self.logger.info(f"Controller released for pipeline '{self.pipeline_id}'")
                except Exception as e:
                    self.logger.error(f"Error releasing controller: {e}")
            
            self.logger.info(f"Checking for child processes for pipeline '{self.pipeline_id}'")
            processes = multiprocessing.active_children()
            if processes:
                self.logger.info(f"Cleaning up {len(processes)} child processes for pipeline '{self.pipeline_id}'")
                for process in processes:
                    if process.is_alive():
                        self.logger.info(f"Terminating child process {process.pid}")
                        try:
                            process.terminate()
                            process.join(timeout=1.0)  
                            if process.is_alive():
                                self.logger.warning(f"Force killing child process {process.pid}")
                                process.kill()
                                process.join(timeout=0.5)
                        except Exception as e:
                            self.logger.error(f"Error killing process {process.pid}: {e}")
            else:
                self.logger.info(f"No child processes found for pipeline '{self.pipeline_id}'")
            
            try:
                if get_broker().stop_stream(str(self.pipeline_id)):
                    self.logger.info(f"Stopped stream for pipeline '{self.pipeline_id}'")
            except Exception as e:
                self.logger.error(f"Error stopping stream for pipeline '{self.pipeline_id}': {e}")
            
            self.state = PipelineState.STOPPED
            self.logger.info(f"Pipeline '{self.pipeline_id}' stopped successfully")
        except Exception as e:
            self.state = PipelineState.ERROR
            self.error = str(e)
            self.logger.error(f"Pipeline '{self.pipeline_id}' failed to stop: {e}")


class PipelineManager:
    def __init__(self):
        self._lock = threading.RLock()
        self._items: Dict[int, PipelineRunner] = {}
        self._shutdown_called = False
        self.logger = get_module_logger("api.pipeline_manager")

    def create(self, pid: int, config_body: dict, name: str = None) -> Dict:
        with self._lock:
            if pid in self._items:
                raise ValueError("Pipeline already exists")
            runner = PipelineRunner(pid, config_body, name)
            self._items[pid] = runner
            self.logger.info(f"Pipeline '{pid}' created with name '{runner.pipeline_name}'")
            return self._describe_locked(pid, runner)

    def delete(self, pid: int) -> Dict:
        with self._lock:
            runner = self._items.get(pid)
            if runner is None:
                raise KeyError("Pipeline not found")
            if runner.state == PipelineState.RUNNING:
                raise RuntimeError("Stop pipeline before delete")
            self._items.pop(pid)
            self.logger.info(f"Pipeline '{pid}' deleted")
            return {"id": pid, "status": "deleted"}

    def start(self, pid: int) -> Dict:
        with self._lock:
            runner = self._items.get(pid)
            if runner is None:
                raise KeyError("Pipeline not found")
        
        runner.start()
        
        with self._lock:
            return {
                "id": pid,
                "name": runner.pipeline_name,
                "state": runner.state,
                "error": runner.error,
            }

    def stop(self, pid: int) -> Dict:
        with self._lock:
            runner = self._items.get(pid)
            if runner is None:
                raise KeyError("Pipeline not found")
        
        runner.stop()
        
        with self._lock:
            return {
                "id": pid,
                "name": runner.pipeline_name,
                "state": runner.state,
                "error": runner.error,
            }

    def list(self) -> Dict[int, Dict]:
        with self._lock:
            return {pid: self._describe_locked(pid, r) for pid, r in self._items.items()}

    def describe(self, pid: int) -> Dict:
        with self._lock:
            runner = self._items.get(pid)
            if runner is None:
                raise KeyError("Pipeline not found")
            return self._describe_locked(pid, runner)

    def _get_runner(self, pid: int) -> Optional[PipelineRunner]:
        """Get runner for a specific pipeline (thread-safe)."""
        with self._lock:
            return self._items.get(pid)
    
    def _describe_locked(self, pid: int, runner: PipelineRunner) -> Dict:
        return {
            "id": pid,
            "name": runner.pipeline_name,
            "state": runner.state,
            "error": runner.error,
        }

    def shutdown(self) -> None:
        with self._lock:
            if self._shutdown_called:
                self.logger.info("PipelineManager shutdown already called, skipping")
                return
            self._shutdown_called = True
        
        self.logger.info("PipelineManager shutdown initiated")
        
        with self._lock:
            ids = list(self._items.keys())
        
        self.logger.info(f"Stopping {len(ids)} pipelines: {ids}")
        
        for pid in ids:
            try:
                self.logger.info(f"Stopping pipeline '{pid}'")
                self.stop(pid)
            except Exception as e:
                self.logger.error(f"Error stopping pipeline '{pid}': {e}")
        
        import time
        time.sleep(1.0)
        
        processes = multiprocessing.active_children()
        if processes:
            self.logger.warning(f"Found {len(processes)} active processes after pipeline stop, force cleaning up...")
            for process in processes:
                if process.is_alive():
                    self.logger.warning(f"Force terminating process {process.pid} ({process.name})")
                    try:
                        process.terminate()
                        process.join(timeout=0.5)
                        if process.is_alive():
                            self.logger.error(f"Process {process.pid} still alive, force killing")
                            process.kill()
                            process.join(timeout=0.2)
                    except Exception as e:
                        self.logger.error(f"Error killing process {process.pid}: {e}")
        
        remaining = multiprocessing.active_children()
        if remaining:
            self.logger.error(f"CRITICAL: Still {len(remaining)} processes alive after cleanup!")
            for process in remaining:
                self.logger.error(f"Process {process.pid} ({process.name}) still alive")
                try:
                    if hasattr(process, 'kill'):
                        process.kill()
                    else:
                        if os.name == 'nt':  
                            subprocess.run(['taskkill', '/F', '/PID', str(process.pid)], 
                                         capture_output=True)
                        else:  
                            os.kill(process.pid, 9)  
                except Exception as e:
                    self.logger.error(f"Failed to kill process {process.pid}: {e}")
        
        self.logger.info("PipelineManager shutdown completed")





import pathlib
import json
import datetime
import numpy as np
import cv2
from ..object_tracker.tracking_results import TrackingResult
from ..objects_handler.object_result import ObjectResultHistory
import copy
from pathlib import Path
from ..core.logger import get_module_logger

# Инициализация логгера для утилит
utils_logger = get_module_logger("utils")

from sympy.multipledispatch.dispatcher import source

from ..database_controller import database_controller_pg
from psycopg2 import sql
from ..capture.video_capture_base import CaptureImage
# from object_tracker.object_tracking_botsort import BOTrack
from ..object_tracker.trackers.sctrack import SCTrack


def get_project_root() -> Path:
    return Path(__file__).parent.parent


def boxes_iou(box1, box2):
    area1 = (box1[2] - box1[0] + 1) * (box1[3] - box1[1] + 1)
    area2 = (box2[2] - box2[0] + 1) * (box2[3] - box2[1] + 1)
    if (((box1[0] <= box2[0] and box1[1] <= box2[1]) and (
            box2[2] <= box1[2] and box2[3] <= box1[3])) or  # Находится ли один bbox внутри другого
            ((box2[0] <= box1[0] and box2[1] <= box1[1]) and (box1[2] <= box2[2] and box1[3] <= box2[3]))):
        return 1.0, box1 if area1 > area2 else box2
    x_left = max(box1[0], box2[0])
    y_top = max(box1[1], box2[1])
    x_right = min(box1[2], box2[2])
    y_bottom = min(box1[3], box2[3])
    if x_right - x_left + 1 <= 0 or y_bottom - y_top + 1 <= 0:  # Если рамки никак не пересекаются
        return -1.0, None
    intersection = (x_right - x_left + 1) * (y_bottom - y_top + 1)
    iou = intersection / float(area1 + area2 - intersection)
    return iou, box1 if area1 > area2 else box2


def non_max_sup(boxes_coords, confidences, class_ids):
    confidences = np.array(confidences, dtype='float64')
    boxes_coords = np.array(boxes_coords, dtype='float64')
    class_ids = np.array(class_ids, dtype='float64')
    sorted_idxs = np.argsort(confidences)
    iou_thresh = 0.5
    keep_idxs = []
    while len(sorted_idxs) > 0:
        last = len(sorted_idxs) - 1
        suppress_idxs = [last]  # Индекс рамки с наибольшей вероятностью
        keep_idxs.append(sorted_idxs[last])
        for i in range(len(sorted_idxs) - 1):
            idx = sorted_idxs[i]
            iou, max_box = boxes_iou(boxes_coords[sorted_idxs[last]], boxes_coords[idx])
            if iou > iou_thresh:  # Если iou превышает порог, то добавляем данную рамку на удаление
                boxes_coords[idx] = copy.deepcopy(max_box)
                suppress_idxs.append(i)
        sorted_idxs = np.delete(sorted_idxs, suppress_idxs)
    boxes_coords = boxes_coords[keep_idxs].tolist()
    class_ids = class_ids[keep_idxs].tolist()
    confidences = confidences[keep_idxs].tolist()
    return boxes_coords, confidences, class_ids


def roi_to_image(roi_box_coords, x0, y0):
    image_box_coords = [x0 + int(roi_box_coords[0]), y0 + int(roi_box_coords[1]),
                        x0 + int(roi_box_coords[2]), y0 + int(roi_box_coords[3])]
    return image_box_coords


def create_roi(capture_image: CaptureImage, coords):
    rois = []
    img = capture_image.image
    if img is None:
        # Return empty list if image is None to avoid TypeError
        return rois
    for x, y, w, h in coords:
        roi_img = img[y:y+h, x:x+w]
        roi_capture = CaptureImage()
        roi_capture.source_id = capture_image.source_id
        roi_capture.frame_id = capture_image.frame_id
        roi_capture.current_video_frame = capture_image.current_video_frame
        roi_capture.current_video_position = capture_image.current_video_position
        roi_capture.time_stamp = capture_image.time_stamp
        roi_capture.image = roi_img
        rois.append([roi_capture, [x, y]])
    return rois


def merge_roi_boxes(all_roi, bboxes_coords, confidences, class_ids):
    bboxes_merged = []
    conf_merged = []
    ids_merged = []
    merged_idxs = []
    for i in range(len(bboxes_coords)):
        intersected_idxs = []
        if i in merged_idxs:
            continue
        for j in range(i + 1, len(bboxes_coords)):
            # Если рамки пересекаются, но находятся в разных регионах, то добавляем их в список пересекающихся
            if ((len(all_roi) != 0) and is_intersected(bboxes_coords[i], bboxes_coords[j])
                    and not is_same_roi(all_roi, bboxes_coords[i], bboxes_coords[j])):
                intersected_idxs.append(j)
        # Если рамка пересекается больше, чем с одной, то проверяем, с какой она пересекается больше, чтобы их объединить
        if len(intersected_idxs) > 1:
            iou = []
            # Определяем, с какой рамкой iou выше
            for k in range(len(intersected_idxs)):
                iou.append(boxes_iou(bboxes_coords[i], bboxes_coords[intersected_idxs[k]]))
            max_idx = iou.index(max(iou))
            # Объединяем с этой рамкой
            bboxes_coords[i] = [min(bboxes_coords[i][0], bboxes_coords[intersected_idxs[max_idx]][0]),
                                min(bboxes_coords[i][1], bboxes_coords[intersected_idxs[max_idx]][1]),
                                max(bboxes_coords[i][2], bboxes_coords[intersected_idxs[max_idx]][2]),
                                max(bboxes_coords[i][3], bboxes_coords[intersected_idxs[max_idx]][3])]
            confidences[i] = max(confidences[i], confidences[intersected_idxs[max_idx]])
            merged_idxs.append(intersected_idxs[max_idx])
        # Если пересекается только с одной, объединяем
        elif len(intersected_idxs) == 1:
            bboxes_coords[i] = [min(bboxes_coords[i][0], bboxes_coords[intersected_idxs[0]][0]),
                                min(bboxes_coords[i][1], bboxes_coords[intersected_idxs[0]][1]),
                                max(bboxes_coords[i][2], bboxes_coords[intersected_idxs[0]][2]),
                                max(bboxes_coords[i][3], bboxes_coords[intersected_idxs[0]][3])]
            confidences[i] = max(confidences[i], confidences[intersected_idxs[0]])
            merged_idxs.append(intersected_idxs[0])
        bboxes_merged.append(bboxes_coords[i])
        conf_merged.append(confidences[i])
        ids_merged.append(class_ids[i])
    return bboxes_merged, conf_merged, ids_merged


def is_same_roi(all_roi, box1, box2):
    if len(all_roi) == 0:
        return True
    surrounding_rois_box1 = []
    surrounding_rois_box2 = []
    for i, roi in enumerate(all_roi):
        if (((roi[1] <= box1[3] <= (roi[1] + roi[3])) and (roi[1] <= box1[1] <= (roi[1] + roi[3]))) and
                ((roi[1] <= box2[3] <= (roi[1] + roi[3])) and (roi[1] <= box2[1] <= (roi[1] + roi[3])))):
            # Если рамки находятся в одном регионе, но хотя бы одна из рамок уже находится в другом, значит
            # регионы вложенные, поэтому возвращаем False и объединяем рамки
            if len(surrounding_rois_box1) > 0 or len(surrounding_rois_box2) > 0:
                return False
            return True
        elif ((roi[1] <= box1[3] <= (roi[1] + roi[3])) and (roi[1] <= box1[1] <= (roi[1] + roi[3])) and not
        ((roi[1] <= box2[3] <= (roi[1] + roi[3])) and (roi[1] <= box2[1] <= (roi[1] + roi[3])))):
            # Проверка на вложенность регионов интереса, создаем для каждой рамки список окружающих регионов
            surrounding_rois_box1.append(i)
        elif ((roi[1] <= box2[3] <= (roi[1] + roi[3])) and (roi[1] <= box2[1] <= (roi[1] + roi[3])) and not
        ((roi[1] <= box1[3] <= (roi[1] + roi[3])) and (roi[1] <= box1[1] <= (roi[1] + roi[3])))):
            # Проверка на вложенность регионов интереса, создаем для каждой рамки список окружающих регионов
            surrounding_rois_box2.append(i)
    return False


def is_intersected(box1, box2):
    if ((box1[2] >= box2[0]) and (box2[2] >= box1[0])) and ((box1[3] >= box2[1]) and (box2[3] >= box1[1])):
        return True
    else:
        return False


def get_objs_info(bboxes_coords, confidences, class_ids):
    objects = []
    for bbox, class_id, conf in zip(bboxes_coords, class_ids, confidences):
        obj = {"bbox": bbox, "conf": conf, "class": class_id}
        objects.append(obj)
    return objects


def get_class_name_from_mapping(class_id: int, class_mapping: dict) -> str:
    """
    Get class name from class ID using class_mapping.
    
    Args:
        class_id: Class ID
        class_mapping: Dictionary mapping class_name -> class_id
        
    Returns:
        Class name string or 'class_{id}' if not found
    """
    for name, cid in class_mapping.items():
        if cid == class_id:
            return name
    return f"class_{class_id}"


def draw_boxes(image, objects, cam_id, class_mapping, text_config=None):
    """
    Draw bounding boxes and labels with adaptive text positioning.
    
    Args:
        image: OpenCV image
        objects: List of detected objects
        cam_id: Camera ID
        class_mapping: Class mapping dict {class_name: class_id}
        text_config: Text configuration dictionary (optional)
    """
    # Apply text configuration
    config = apply_text_config(text_config)
    
    for cam_objs in objects:
        if cam_objs['cam_id'] == cam_id:
            for obj in cam_objs['objects']:
                # Draw bounding box
                cv2.rectangle(image, (int(obj['bbox'][0]), int(obj['bbox'][1])),
                              (int(obj['bbox'][2]), int(obj['bbox'][3])), (0, 255, 0), thickness=8)
                
                # Get class name from class_mapping
                class_name = get_class_name_from_mapping(obj['class'], class_mapping)
                
                # Create text label
                text = str(class_name) + " " + "{:.2f}".format(obj['conf'])
                
                # Draw text with adaptive positioning
                put_text_with_bbox(image, text, obj['bbox'], 
                                 font_size_pt=config['font_size_pt'],
                                 font_face=config['font_face'],
                                 color=config['color'],
                                 thickness=config['thickness'],
                                 background_color=config['background_color'],
                                 position_offset_percent=config['position_offset_percent'],
                                 font_scale_method=config.get('font_scale_method', 'resolution_based'),
                                 base_resolution=config.get('base_resolution', (1920, 1080)),
                                 background_enabled=config.get('background_enabled', True))


def draw_preview_boxes(image, width, height, box):
    cv2.rectangle(image, (int(box[0] * width), int(box[1] * height)),
                  (int(box[2] * width), int(box[3] * height)), (0, 255, 0), thickness=1)
    return image


def draw_preview_boxes_zones(image, width, height, box, zone_coords):
    points = [(int(point[0] * width), int(point[1] * height)) for point in zone_coords]
    image = cv2.rectangle(image, (int(box[0] * width), int(box[1] * height)),
                          (int(box[2] * width), int(box[3] * height)), (0, 255, 0), thickness=1)
    image = cv2.polylines(image, pts=np.int32([points]), isClosed=True, color=(0, 0, 255), thickness=1)
    return image


def draw_boxes_from_db(db_controller, table_name, load_folder, save_folder):
    query = sql.SQL(
        'SELECT object_id, confidence, bounding_box, lost_bounding_box, frame_path, lost_frame_path FROM {table};').format(
        table=sql.Identifier(table_name))
    res = db_controller.query(query)
    for obj_id, conf, box, lost_box, image_path, lost_image_path in res:
        lost_load_dir = pathlib.Path(load_folder, 'lost')
        detected_load_dir = pathlib.Path(load_folder, 'detected')

        if not lost_load_dir.exists():
            Path.mkdir(lost_load_dir)
        lost_load_path = pathlib.Path(lost_load_dir, Path(lost_image_path).name)
        if not detected_load_dir.exists():
            Path.mkdir(detected_load_dir)
        detected_load_path = pathlib.Path(detected_load_dir, Path(image_path).name)

        if not save_folder.exists():
            pathlib.Path.mkdir(save_folder)

        lost_save_dir = pathlib.Path(save_folder, 'lost')
        if not lost_save_dir.exists():
            Path.mkdir(lost_save_dir)
        detected_save_dir = pathlib.Path(save_folder, 'detected')
        if not detected_save_dir.exists():
            Path.mkdir(detected_save_dir)
        lost_save_path = pathlib.Path(lost_save_dir, Path(lost_image_path).name)
        detected_save_path = pathlib.Path(detected_save_dir, Path(image_path).name)

        lost_image = cv2.imread(lost_load_path.as_posix())
        cv2.rectangle(lost_image, (int(box[0]), int(box[1])),
                      (int(box[2]), int(box[3])), (0, 255, 0), thickness=8)
        cv2.putText(lost_image, str(obj_id) + " " + "{:.2f}".format(conf),
                    (int(box[0]), int(box[1]) - 10), cv2.FONT_HERSHEY_SIMPLEX, 1,
                    (255, 255, 255), 2)

        detected_image = cv2.imread(detected_load_path.as_posix())
        cv2.rectangle(detected_image, (int(box[0]), int(box[1])),
                      (int(box[2]), int(box[3])), (0, 255, 0), thickness=8)
        cv2.putText(detected_image, str(obj_id) + " " + "{:.2f}".format(conf),
                    (int(box[0]), int(box[1]) - 10), cv2.FONT_HERSHEY_SIMPLEX, 1,
                    (255, 255, 255), 2)
        lost_saved = cv2.imwrite(lost_save_path.as_posix(), lost_image)
        detected_saved = cv2.imwrite(detected_save_path.as_posix(), detected_image)
        if not lost_saved or not detected_saved:
            utils_logger.error('Error saving image with boxes')


def draw_boxes_tracking(image: CaptureImage, cameras_objs, source_name, source_duration_msecs, font_scale, font_thickness, font_color, text_config=None, class_mapping=None, event_active_obj_ids=None, event_color=(0, 0, 255)):
    height, width, channels = image.image.shape
    
    # Apply text configuration
    config = apply_text_config(text_config)
    
    # Draw source name
    if source_name is int:
        source_text = "Source Id: " + str(source_name)
    else:
        source_text = str(source_name)
    
    # Position source name at bottom-left (10% from left, 10% from bottom)
    put_text_adaptive(image.image, source_text, (10, 90), 
                     font_size_pt=config['font_size_pt'],
                     font_face=config['font_face'],
                     color=config['color'],
                     thickness=config['thickness'],
                     background_color=config['background_color'],
                     font_scale_method=config.get('font_scale_method', 'resolution_based'),
                     base_resolution=config.get('base_resolution', (1920, 1080)),
                     background_enabled=config.get('background_enabled', True))

    # Draw time position
    if image.current_video_position and source_duration_msecs is not None:
        time_position_secs = image.current_video_position / 1000.0
        pos_string = "{:.1f}".format(time_position_secs) + " [" + "{:.1f}".format(source_duration_msecs / 1000.0) + "]"
        
        # Position time at bottom-right (10% from right, 10% from bottom)
        put_text_adaptive(image.image, pos_string, (90, 90), 
                         font_size_pt=config['font_size_pt'],
                         font_face=config['font_face'],
                         color=config['color'],
                         thickness=config['thickness'],
                         background_color=config['background_color'],
                         font_scale_method=config.get('font_scale_method', 'resolution_based'),
                         base_resolution=config.get('base_resolution', (1920, 1080)),
                         background_enabled=config.get('background_enabled', True))

    # Для трекинга отображаем только последние данные об объекте из истории
    # utils_logger.error(cameras_objs)
    for obj in cameras_objs:
        # if obj.frame_id < image.frame_id:
        #     continue

        last_hist_index = len(obj.history) - 1
        last_info = obj.track
        if obj.frame_id != image.frame_id:
            for i in range(len(obj.history) - 1):
                if obj.history[i].frame_id == image.frame_id:
                    last_hist_index = i
                    last_info = obj.history[i].track
                    break

        cv2.rectangle(image.image, (int(last_info.bounding_box[0]), int(last_info.bounding_box[1])),
                      (int(last_info.bounding_box[2]), int(last_info.bounding_box[3])), (0, 255, 0), thickness=font_thickness)

        # Если объект активен по событию — поверх рисуем такой же bbox цветом события
        try:
            if event_active_obj_ids:
                oid = getattr(obj, 'object_id', None)
                if oid is not None and oid in event_active_obj_ids:
                    cv2.rectangle(image.image,
                                  (int(last_info.bounding_box[0]), int(last_info.bounding_box[1])),
                                  (int(last_info.bounding_box[2]), int(last_info.bounding_box[3])),
                                  event_color, thickness=max(2, int(font_thickness * 1.5)))
        except Exception:
            pass
        
        # Create tracking text with class name instead of class_id
        if class_mapping:
            class_name = get_class_name_from_mapping(last_info.class_id, class_mapping)
        else:
            class_name = f"class_{last_info.class_id}"
            
        if obj.global_id is not None:
            tracking_text = f'G{obj.global_id} {class_name} [{last_info.track_id}:{"{:.2f}".format(last_info.confidence)}]'
        else:
            tracking_text = f'{class_name} [{last_info.track_id}:{"{:.2f}".format(last_info.confidence)}]'

        # Calculate font scale based on image resolution
        font_scale_method = config.get('font_scale_method', 'resolution_based')
        font_size_pt = config['font_size_pt']
        base_resolution = config.get('base_resolution', (1920, 1080))
        thickness = config['thickness']
        font_face = config['font_face']

        if font_scale_method == "simple":
            font_scale = calculate_font_scale_simple(font_size_pt, width, height)
        else:  # resolution_based
            font_scale = calculate_font_scale_for_resolution(font_size_pt, width, height, base_resolution)

        # Auto-calculate thickness if not provided
        if thickness is None:
            thickness = max(1, int(font_scale * 2))

        # Draw tracking text with adaptive positioning
        put_text_with_bbox(image.image, tracking_text, last_info.bounding_box,
                          font_face=font_face,
                          font_scale=font_scale,
                          color=config['color'],
                          thickness=thickness,
                          background_color=config['background_color'],
                          position_offset_percent=config['position_offset_percent'],
                          background_enabled=config.get('background_enabled', True))
        
        # Draw attributes if available
        if hasattr(obj, 'attributes') and obj.attributes:
            draw_object_attributes(image.image, obj, last_info.bounding_box, font_face, font_scale*0.5, thickness)

        # utils_logger.error(len(obj['obj_info']))
        if len(obj.history) > 1:
            for i in range(0, last_hist_index):
                first_info = obj.history[i].track
                second_info = obj.history[i + 1].track
                first_cm_x = int((first_info.bounding_box[0] + first_info.bounding_box[2]) / 2)
                first_cm_y = int(first_info.bounding_box[3])
                second_cm_x = int((second_info.bounding_box[0] + second_info.bounding_box[2]) / 2)
                second_cm_y = int(second_info.bounding_box[3])
                cv2.line(image.image, (first_cm_x, first_cm_y),
                         (second_cm_x, second_cm_y), (0, 0, 255), thickness=font_thickness)


def draw_debug_info(image: CaptureImage, debug_info: dict):
    if not debug_info:
        return
    if 'detectors' not in debug_info.keys():
        return

    for det_id, det_debug_info in debug_info['detectors'].items():
        if 'source_ids' in det_debug_info.keys() and image.source_id in det_debug_info['source_ids']:
            source_id_index = det_debug_info['source_ids'].index(image.source_id)
            rois = det_debug_info['roi']
            if type(rois) is list and source_id_index in range(len(rois)):
                for roi in rois[source_id_index]:
                    cv2.rectangle(image.image, (int(roi[0]), int(roi[1])),
                                  (int(roi[0] + roi[2]), int(roi[1] + roi[3])), (255, 0, 0), thickness=9)


class ObjectResultEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        if isinstance(obj, TrackingResult):
            return obj.__dict__
        if isinstance(obj, ObjectResultHistory):
            return obj.__dict__
        if isinstance(obj, CaptureImage):
            return None
        # if isinstance(obj, BOTrack):
        #     return None
        if isinstance(obj, SCTrack):
            return None

        return super().default(obj)


def normalize_config_path(config_path):
    """
    Normalize configuration file path by adding 'configs/' prefix if not present.
    
    Args:
        config_path: Path to configuration file (string or Path object)
        
    Returns:
        Normalized path as string
        
    Examples:
        >>> normalize_config_path("my_config.json")
        "configs/my_config.json"
        >>> normalize_config_path("configs/existing.json")
        "configs/existing.json"
        >>> normalize_config_path("/absolute/path/config.json")
        "/absolute/path/config.json"
    """
    import os
    from pathlib import Path
    
    config_path_str = str(config_path)
    
    # If it's already an absolute path or already has configs/ prefix, return as is
    if os.path.isabs(config_path_str) or config_path_str.startswith("configs"):
        return config_path_str
    
    # Add configs/ prefix
    return os.path.join("configs", config_path_str)


# =============================================================================
# TEXT RENDERING UTILITIES
# =============================================================================

def pt_to_pixels(pt_size, dpi=96):
    """
    Convert points to pixels based on DPI.
    
    Args:
        pt_size: Font size in points
        dpi: Dots per inch (default: 96 for standard screen)
        
    Returns:
        Font size in pixels
    """
    return int(pt_size * dpi / 72.0)


def calculate_font_scale_for_resolution(font_size_pt, image_width, image_height, base_resolution=(1920, 1080)):
    """
    Calculate font scale based on image resolution.
    
    This function calculates an appropriate font scale for OpenCV's putText function
    based on the desired font size in points and the current image resolution.
    
    Args:
        font_size_pt: Desired font size in points
        image_width: Width of the image in pixels
        image_height: Height of the image in pixels
        base_resolution: Base resolution for scaling (default: 1920x1080)
        
    Returns:
        Font scale value for OpenCV putText function
    """
    # Convert points to pixels at 96 DPI
    font_size_px = pt_to_pixels(font_size_pt, dpi=96)
    
    # Calculate resolution factor based on image size
    # Use the smaller dimension to avoid oversized text on very wide/tall images
    image_min_dimension = min(image_width, image_height)
    base_min_dimension = min(base_resolution[0], base_resolution[1])
    
    # Calculate scaling factor
    resolution_factor = image_min_dimension / base_min_dimension
    
    # Apply resolution scaling to font size
    scaled_font_size_px = font_size_px * resolution_factor
    
    # Convert to OpenCV font scale (approximate conversion)
    # OpenCV font scale is roughly pixels / 30 for FONT_HERSHEY_SIMPLEX
    font_scale = scaled_font_size_px / 30.0
    
    # Ensure minimum and maximum reasonable values
    font_scale = max(0.1, min(font_scale, 10.0))
    
    return font_scale


def calculate_font_scale_simple(font_size_pt, image_width, image_height):
    """
    Simple font scale calculation based on image diagonal.
    
    Args:
        font_size_pt: Desired font size in points
        image_width: Width of the image in pixels
        image_height: Height of the image in pixels
        
    Returns:
        Font scale value for OpenCV putText function
    """
    # Convert points to pixels
    font_size_px = pt_to_pixels(font_size_pt, dpi=96)
    
    # Calculate image diagonal
    image_diagonal = (image_width ** 2 + image_height ** 2) ** 0.5
    
    # Base diagonal for 1920x1080
    base_diagonal = (1920 ** 2 + 1080 ** 2) ** 0.5
    
    # Calculate scaling factor
    scale_factor = image_diagonal / base_diagonal
    
    # Apply scaling and convert to OpenCV scale
    scaled_font_size_px = font_size_px * scale_factor
    font_scale = scaled_font_size_px / 30.0
    
    # Ensure reasonable bounds
    font_scale = max(0.1, min(font_scale, 10.0))
    
    return font_scale


def percent_to_pixels(percent, total_size):
    """
    Convert percentage to pixels.
    
    Args:
        percent: Percentage value (0.0 to 100.0)
        total_size: Total size in pixels
        
    Returns:
        Position in pixels
    """
    return int(percent * total_size / 100.0)


def calculate_text_size(text, font_face=cv2.FONT_HERSHEY_SIMPLEX, font_scale=1.0, thickness=1):
    """
    Calculate text size in pixels.
    
    Args:
        text: Text string
        font_face: OpenCV font face
        font_scale: Font scale
        thickness: Font thickness
        
    Returns:
        Tuple of (width, height) in pixels
    """
    (text_width, text_height), baseline = cv2.getTextSize(text, font_face, font_scale, thickness)
    return text_width, text_height + baseline


def get_adaptive_font_scale(text, target_width_px, font_face=cv2.FONT_HERSHEY_SIMPLEX, thickness=1):
    """
    Calculate font scale to fit text within target width.
    
    Args:
        text: Text string
        target_width_px: Target width in pixels
        font_face: OpenCV font face
        thickness: Font thickness
        
    Returns:
        Font scale that fits the text
    """
    # Start with scale 1.0 and reduce until text fits
    scale = 1.0
    while scale > 0.1:  # Minimum scale limit
        text_width, _ = calculate_text_size(text, font_face, scale, thickness)
        if text_width <= target_width_px:
            return scale
        scale *= 0.9
    return 0.1


def put_text_adaptive(image, text, position_percent, font_size_pt=12, font_face=cv2.FONT_HERSHEY_SIMPLEX, 
                     color=(255, 255, 255), thickness=None, background_color=None, padding_percent=2.0,
                     font_scale_method="resolution_based", base_resolution=(1920, 1080), background_enabled=True):
    """
    Draw text with adaptive positioning and sizing.
    
    Args:
        image: OpenCV image
        text: Text to draw
        position_percent: Position as (x_percent, y_percent) from top-left
        font_size_pt: Font size in points
        font_face: OpenCV font face
        color: Text color (B, G, R)
        thickness: Font thickness (auto-calculated if None)
        background_color: Background color for text (optional)
        padding_percent: Padding around text in percent of image width
        font_scale_method: Method for calculating font scale ("resolution_based" or "simple")
        base_resolution: Base resolution for scaling (width, height)
        background_enabled: Whether to draw background rectangle (True/False)
        
    Returns:
        Modified image
    """
    height, width = image.shape[:2]
    
    # Calculate font scale based on image resolution
    if font_scale_method == "simple":
        font_scale = calculate_font_scale_simple(font_size_pt, width, height)
    else:  # resolution_based
        font_scale = calculate_font_scale_for_resolution(font_size_pt, width, height, base_resolution)
    
    # Auto-calculate thickness if not provided
    if thickness is None:
        thickness = max(1, int(font_scale * 2))
    
    # Calculate position in pixels
    x_px = percent_to_pixels(position_percent[0], width)
    y_px = percent_to_pixels(position_percent[1], height)
    
    # Calculate text size
    text_width, text_height = calculate_text_size(text, font_face, font_scale, thickness)
    
    # Adjust position to ensure text fits within image bounds
    if x_px + text_width > width:
        x_px = width - text_width - 10
    if y_px - text_height < 0:
        y_px = text_height + 10
    
    # Draw background if specified and enabled
    if background_color and background_enabled:
        padding_px = percent_to_pixels(padding_percent, width)
        cv2.rectangle(image, 
                     (x_px - padding_px, y_px - text_height - padding_px),
                     (x_px + text_width + padding_px, y_px + padding_px),
                     background_color, -1)
    
    # Draw text
    cv2.putText(image, text, (x_px, y_px), font_face, font_scale, color, thickness)
    
    return image


def put_text_with_bbox(image, text, bbox, font_face, font_scale, thickness,
                      color=(255, 255, 255), background_color=None,
                      position_offset_percent=(0, -10), background_enabled=True):
    """
    Draw text near a bounding box with adaptive positioning.
    
    Args:
        image: OpenCV image
        text: Text to draw
        bbox: Bounding box (x1, y1, x2, y2)
        font_size_pt: Font size in points
        font_face: OpenCV font face
        color: Text color (B, G, R)
        thickness: Font thickness (auto-calculated if None)
        background_color: Background color for text (optional)
        position_offset_percent: Offset from bbox in percent of image size
        font_scale_method: Method for calculating font scale ("resolution_based" or "simple")
        base_resolution: Base resolution for scaling (width, height)
        background_enabled: Whether to draw background rectangle (True/False)
        
    Returns:
        Modified image
    """
    height, width = image.shape[:2]
    
    # Calculate position relative to bbox
    x_offset = percent_to_pixels(position_offset_percent[0], width)
    y_offset = percent_to_pixels(position_offset_percent[1], height)
    
    x_px = int(bbox[0]) + x_offset
    y_px = int(bbox[1]) + y_offset
    
    # Ensure text doesn't go outside image bounds
    if x_px < 0:
        x_px = 10
    if y_px < 0:
        y_px = 10
    
    # Calculate text size
    text_width, text_height = calculate_text_size(text, font_face, font_scale, thickness)
    
    # Adjust if text would go outside image
    if x_px + text_width > width:
        x_px = width - text_width - 10
    if y_px - text_height < 0:
        y_px = text_height + 10
    
    # Draw background if specified and enabled
    if background_color and background_enabled:
        padding_px = 5
        cv2.rectangle(image, 
                     (x_px - padding_px, y_px - text_height - padding_px),
                     (x_px + text_width + padding_px, y_px + padding_px),
                     background_color, -1)
    
    # Draw text
    cv2.putText(image, text, (x_px, y_px), font_face, font_scale, color, thickness)
    
    return image


def draw_object_attributes(image, obj, bbox, font_face, font_scale, thickness):
    """
    Draw object attributes as colored indicators.
    
    Args:
        image: OpenCV image
        obj: ObjectResult object with attributes
        bbox: Bounding box coordinates [x1, y1, x2, y2]
        config: Text configuration dictionary
        font_scale: Font scale from main drawing function (optional)
    """
    if not hasattr(obj, 'attributes') or not obj.attributes:
        return
    
    # Position attributes below the main label
    x1, y1, x2, y2 = bbox
    attr_y = y2 + 5  # Start below the bounding box
    
    # Color mapping for attribute states
    state_colors = {
        'none': (128, 128, 128),    # Gray
        'exists': (0, 255, 0),      # Green  
        'lost': (0, 255, 255)       # Yellow
    }
    
    # Calculate font scale for attributes (0.5x of object class font scale)
    attr_font_scale = font_scale
    attr_thickness = thickness
    attr_font_face = font_face

    attr_texts = []
    for attr_name, attr_data in obj.attributes.items():
        if isinstance(attr_data, dict):
            state = attr_data.get('state', 'none')
            confidence = attr_data.get('confidence_smooth', 0.0)
            total_time = attr_data.get('total_time_ms', 0)
            
            # Получаем новые поля для улучшенного отображения
            total_found_time = attr_data.get('total_found_time_ms', 0)
            total_lost_time = attr_data.get('total_lost_time_ms', 0)
            found_ratio = attr_data.get('found_ratio', 0.0)
            
            # Рассчитываем суммарное время (или ноль если < 0)
            summary_time = max(0, total_found_time - total_lost_time)
            
            # Create attribute text with enhanced information
            color = state_colors.get(state, (128, 128, 128))
            attr_text = f"{attr_name}: {state} ({confidence:.2f}, {summary_time}ms, {found_ratio:.1%})"
            attr_texts.append((attr_text, color))
    
    # Calculate text height for proper spacing
    if attr_texts:
        # Use first attribute text to calculate height
        sample_text = attr_texts[0][0]
        (_, text_height), baseline = cv2.getTextSize(sample_text, attr_font_face,
                                                    attr_font_scale, attr_thickness)
        line_spacing = text_height + 4  # Add 4 pixels padding between lines
    else:
        line_spacing = 20  # Fallback spacing
    
    # Draw attribute texts
    for i, (attr_text, color) in enumerate(attr_texts):
        text_y = attr_y + i * line_spacing
        
        # Draw background rectangle for better visibility
        (text_width, text_height), baseline = cv2.getTextSize(attr_text, attr_font_face,
                                                             attr_font_scale, attr_thickness)
        # Ensure coordinates are integers
        rect_x1 = int(x1)
        rect_y1 = int(text_y - text_height - 2)
        rect_x2 = int(x1 + text_width + 4)
        rect_y2 = int(text_y + 2)
        cv2.rectangle(image, (rect_x1, rect_y1), (rect_x2, rect_y2), (0, 0, 0), -1)
        
        # Draw attribute text
        cv2.putText(image, attr_text, (int(x1 + 2), int(text_y)), 
                   attr_font_face, attr_font_scale,
                   color, attr_thickness)


def get_default_text_config():
    """
    Get default text configuration.
    
    Returns:
        Dictionary with default text settings
    """
    return {
        "font_size_pt": 42,
        "font_face": cv2.FONT_HERSHEY_SIMPLEX,
        "color": (0, 0, 255),  # White
        "thickness": None,  # Auto-calculated
        "background_color": (0, 0, 0),  # No background
        "background_enabled": False,  # Enable/disable background
        "padding_percent": 2.0,
        "position_offset_percent": (0, -10),
        "font_scale_method": "resolution_based",  # "resolution_based" or "simple"
        "base_resolution": (1920, 1080)  # Base resolution for scaling
    }


def apply_text_config(text_config, default_config=None):
    """
    Apply text configuration with defaults.
    
    Args:
        text_config: User-provided text configuration
        default_config: Default configuration (optional)
        
    Returns:
        Merged configuration
    """
    if default_config is None:
        default_config = get_default_text_config()
    
    merged_config = default_config.copy()
    if text_config:
        merged_config.update(text_config)
    
    return merged_config



import os
from fastapi import FastAPI, Response
from fastapi.middleware.cors import CORSMiddleware
from contextlib import asynccontextmanager
from evileye.core.logger import get_module_logger
from evileye.api.routes.configs import router as configs_router
from evileye.api.routes.pipelines import router as pipelines_router
from evileye.api.routes.streaming import router as streaming_router
from evileye.api.routes.events import router as events_router
from evileye.api.core.config_run_access import get_config_run_manager
from evileye.api.core.manager_access import get_manager
from evileye import __version__

logger = get_module_logger("api.app")

@asynccontextmanager
async def lifespan(_app: FastAPI):
    logger.info("FastAPI lifespan startup")
    
    try:
        yield
    finally:
        logger.info("FastAPI lifespan shutdown")
        try:
            get_config_run_manager().shutdown()
        except Exception as e:
            logger.error(f"Error during ConfigRunManager shutdown: {e}")


def create_app() -> FastAPI:
    app = FastAPI(title="EvilEye API", version=os.getenv("EVILEYE_API_VERSION", "v1"), lifespan=lifespan)
    logger.info("FastAPI app created")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=os.getenv("EVILEYE_CORS_ALLOW_ORIGINS", "*").split(","),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    logger.info("CORS middleware configured")

    @app.get("/ready")
    async def ready():
        return {"status": "ok"}

    @app.get("/api/v1/version")
    async def version():
        return {"evileye": __version__, "api": app.version}

    app.include_router(configs_router)
    app.include_router(pipelines_router)
    app.include_router(streaming_router)
    app.include_router(events_router)
    logger.info("Routers registered: configs, pipelines, streaming, events")

    return app



import time
from .db_adapter import DatabaseAdapterBase
from ..utils import threading_events
from ..utils import utils
import os
import copy
import cv2
from psycopg2 import sql


class DatabaseAdapterAttributeEvents(DatabaseAdapterBase):
    def __init__(self, db_controller):
        super().__init__(db_controller)

    def set_params_impl(self):
        super().set_params_impl()
        # event_name must match AttributeEvent.get_name()
        self.event_name = self.params['event_name']

    def _insert_impl(self, event):
        fields, data, preview_path, frame_path = self._prepare_for_saving(event)
        query_type = 'insert'
        insert_query = sql.SQL("INSERT INTO {} ({}) VALUES ({})").format(
            sql.Identifier(self.table_name),
            sql.SQL(",").join(map(sql.Identifier, fields)),
            sql.SQL(', ').join(sql.Placeholder() * len(fields))
        )
        self.queue_in.put((query_type, insert_query, data, preview_path, frame_path, getattr(event, 'img_found', None), getattr(event, 'box_found', None)))

    def _update_impl(self, event):
        fields, data, preview_path, frame_path = self._prepare_for_updating(event)
        query_type = 'update'
        data.append(event.event_id)
        data = tuple(data)
        update_query = sql.SQL('UPDATE {table} SET {data} WHERE event_id=({selected})').format(
            table=sql.Identifier(self.table_name),
            data=sql.SQL(', ').join(
                sql.Composed([sql.Identifier(field), sql.SQL(" = "), sql.Placeholder()]) for field in fields),
            selected=sql.Placeholder(),
            fields=sql.SQL(",").join(map(sql.Identifier, fields)))
        self.queue_in.put((query_type, update_query, data, preview_path, frame_path, getattr(event, 'img_finished', None), getattr(event, 'box_finished', None)))

    def _execute_query(self):
        while self.run_flag:
            time.sleep(0.01)
            try:
                if not self.queue_in.empty():
                    query_type, query_string, data, preview_path, frame_path, image, box = self.queue_in.get()
                    if query_string is not None:
                        pass
                else:
                    query_type = query_string = data = preview_path = frame_path = image = box = None
            except ValueError:
                break

            if query_string is None:
                continue

            try:
                self.db_controller.query(query_string, data)
            except Exception as e:
                # Attempt to auto-migrate missing columns and retry once
                msg = str(e)
                if 'UndefinedColumn' in msg or 'does not exist' in msg:
                    try:
                        self._ensure_attribute_columns()
                        self.logger.warning('DB: Missing columns in attribute_events detected. Applied auto-migration. Retrying query...')
                        self.db_controller.query(query_string, data)
                    except Exception as e2:
                        self.logger.error(f'DB: AttributeEvents query failed after migration attempt: {e2}')
                        continue
                else:
                    self.logger.error(f'DB: AttributeEvents query failed: {e}')
                    continue
            # Save images if available
            try:
                if image is not None and preview_path is not None and frame_path is not None and box is not None:
                    self._save_image(preview_path, frame_path, image, box)
            except Exception:
                pass
            # Log success write - removed to reduce log noise
            if query_type == 'insert':
                threading_events.notify('new event')
            elif query_type == 'update':
                threading_events.notify('update event')

    def _prepare_for_saving(self, event) -> tuple[list, list, str, str]:
        # Payload with image paths for attribute event START
        fields_for_saving = {
            'event_id': event.event_id,
            'source_id': event.source_id,
            'time_stamp': event.timestamp,
            'time_finished': event.get_time_finished(),
            'object_id': event.object_id,
            'event_name': event.matched_event_name,
            'attrs': ','.join(event.matched_attrs),
            'class_id': event.class_id if event.class_id is not None else -1,
            'box_found': None,
            'box_finished': None,
            'preview_path_found': '',
            'frame_path_found': '',
            'video_path_found': getattr(event, 'video_path_found', None),
            'video_path_finished': None
        }
        # Normalize and set box_found if available
        if event.box_found is not None and event.img_found is not None and hasattr(event.img_found, 'image'):
            ih, iw, _ = event.img_found.image.shape
            box = list(event.box_found)
            norm_box = [box[0]/iw, box[1]/ih, box[2]/iw, box[3]/ih]
            fields_for_saving['box_found'] = norm_box
        preview_path = self._get_img_path('preview', 'attribute_found', event, event.time_found)
        frame_path = self._get_img_path('frame', 'attribute_found', event, event.time_found)
        fields_for_saving['preview_path_found'] = preview_path
        fields_for_saving['frame_path_found'] = frame_path
        return (list(fields_for_saving.keys()), list(fields_for_saving.values()), preview_path, frame_path)

    def _prepare_for_updating(self, event):
        # Update finish timestamp and image paths on event completion
        fields_for_updating = {
            'time_finished': event.get_time_finished(),
            'box_finished': None,
            'preview_path_finished': self._get_img_path('preview', 'attribute_finished', event, time_lost=event.get_time_finished()),
            'frame_path_finished': self._get_img_path('frame', 'attribute_finished', event, time_lost=event.get_time_finished()),
            'video_path_finished': getattr(event, 'video_path_finished', None)
        }
        if event.box_finished is not None and event.img_finished is not None and hasattr(event.img_finished, 'image'):
            ih, iw, _ = event.img_finished.image.shape
            box = list(event.box_finished)
            norm_box = [box[0]/iw, box[1]/ih, box[2]/iw, box[3]/ih]
            fields_for_updating['box_finished'] = norm_box
        return (list(fields_for_updating.keys()), list(fields_for_updating.values()), fields_for_updating['preview_path_finished'], fields_for_updating['frame_path_finished'])

    def _save_image(self, preview_path, frame_path, image, box):
        save_dir = self.db_params['image_dir']
        preview_save_dir = os.path.join(save_dir, preview_path)
        frame_save_dir = os.path.join(save_dir, frame_path)
        # Build preview
        try:
            preview_width = self.db_params.get('preview_width', 300)
            preview_height = self.db_params.get('preview_height', 150)
            # Save original (no debug overlays) for both preview and frame
            preview = cv2.resize(copy.deepcopy(image.image), (preview_width, preview_height), cv2.INTER_NEAREST)
            os.makedirs(os.path.dirname(preview_save_dir), exist_ok=True)
            os.makedirs(os.path.dirname(frame_save_dir), exist_ok=True)
            cv2.imwrite(preview_save_dir, preview)
            cv2.imwrite(frame_save_dir, image.image)
        except Exception as e:
            self.logger.error(f"Attribute event image saving error: {e}")

    def _get_img_path(self, image_type, obj_event_type, event, time_stamp=None, time_lost=None):
        save_dir = self.db_params['image_dir']
        events_dir = os.path.join(save_dir, 'Events')
        import datetime
        cur_date = datetime.date.today()
        cur_date_str = cur_date.strftime('%Y-%m-%d')

        current_day_path = os.path.join(events_dir, cur_date_str)
        images_dir = os.path.join(current_day_path, 'Images')
        # New folders for events: FoundFrames/FoundPreviews/LostFrames/LostPreviews
        if obj_event_type == 'attribute_found':
            if image_type == 'preview':
                subdir = 'FoundPreviews'
            else:
                subdir = 'FoundFrames'
        else:  # attribute_finished
            if image_type == 'preview':
                subdir = 'LostPreviews'
            else:
                subdir = 'LostFrames'
        obj_type_path = os.path.join(images_dir, subdir)

        if not os.path.exists(events_dir):
            os.makedirs(events_dir, exist_ok=True)
        if not os.path.exists(current_day_path):
            os.makedirs(current_day_path, exist_ok=True)
        if not os.path.exists(images_dir):
            os.makedirs(images_dir, exist_ok=True)
        if not os.path.exists(obj_type_path):
            os.makedirs(obj_type_path, exist_ok=True)

        obj_id = event.object_id
        if obj_event_type == 'attribute_found':
            timestamp = (time_stamp or event.timestamp).strftime('%Y-%m-%d_%H-%M-%S.%f')
            img_path = os.path.join(obj_type_path, f'{timestamp}_attr_{event.matched_event_name}_obj{obj_id}_{image_type}.jpeg')
        elif obj_event_type == 'attribute_finished':
            ts = (time_lost or event.get_time_finished())
            timestamp = ts.strftime('%Y-%m-%d_%H-%M-%S-%f') if ts else 'unknown'
            img_path = os.path.join(obj_type_path, f'{timestamp}_attr_{event.matched_event_name}_obj{obj_id}_{image_type}.jpeg')
        return os.path.relpath(img_path, save_dir)

    def _ensure_attribute_columns(self):
        # Ensure newly added columns exist in attribute_events table
        try:
            alter_tpl = "ALTER TABLE {} ADD COLUMN IF NOT EXISTS {} {};"
            table = self.table_name
            # List of required columns and types
            required = [
                ('preview_path_found', 'text'),
                ('frame_path_found', 'text'),
                ('preview_path_finished', 'text'),
                ('frame_path_finished', 'text'),
                ('video_path_found', 'text'),
                ('video_path_finished', 'text'),
                ('class_id', 'integer'),
                ('box_found', 'real[]'),
                ('box_finished', 'real[]')
            ]
            for col, coltype in required:
                query = alter_tpl.format(table, col, coltype)
                self.db_controller.query(query, None)
        except Exception as e:
            self.logger.error(f'DB: Failed to ensure attribute_events columns: {e}')





import json
import os
import sys
from pathlib import Path
from typing import Optional

try:
    from PyQt6.QtWidgets import QApplication, QMessageBox
    from PyQt6.QtCore import QEventLoop, Qt, QTimer
    pyqt_version = 6
except ImportError:
    from PyQt5.QtWidgets import QApplication, QMessageBox
    from PyQt5.QtCore import QEventLoop, Qt, QTimer
    pyqt_version = 5

from evileye.controller import controller
from evileye.visualization_modules.main_window import MainWindow
from evileye.visualization_modules.startup_progress_window import StartupProgressWindow
from evileye.visualization_modules.controller_init_thread import ControllerInitThread
from evileye.utils.utils import normalize_config_path
from evileye.core.logger import get_module_logger
from evileye.gui import GUIManager, GUIMode, determine_gui_mode


def run_config(config_path: str, gui: bool = True, autoclose: bool = False) -> int:
    """Run EvilEye configuration using provided configuration.

    Args:
        config_path: Path to configuration file
        gui: CLI flag for GUI (True/False/None). If None, determined from config.
        autoclose: Auto-close when video ends
        
    Returns:
        Qt application exit code (0 for success)
    """
    logger = get_module_logger("run_config")

    config_file_name = normalize_config_path(config_path)
    config_dir = os.path.dirname(os.path.abspath(config_file_name))
    if config_dir:
        if os.path.basename(config_dir) == 'configs':
            parent_dir = os.path.dirname(config_dir)
            os.chdir(parent_dir)
            logger.info(f"Changed working directory to parent of configs folder: {parent_dir}")
        else:
            os.chdir(config_dir)
            logger.info(f"Changed working directory to: {config_dir}")

    with open(config_file_name) as config_file:
        config_data = json.load(config_file)

    logger.info("Configuration loaded successfully")

    # Ensure controller section exists
    controller_cfg = config_data.setdefault("controller", {})

    # Scheduled restart defaults (used by CLI-level scheduler and potential future GUI integration)
    sched_cfg = controller_cfg.setdefault("scheduled_restart", {})
    sched_cfg.setdefault("enabled", False)
    sched_cfg.setdefault("mode", "daily_time")          # daily_time | interval
    sched_cfg.setdefault("time", "01:00")               # HH:MM, local time
    sched_cfg.setdefault("interval_minutes", 0)         # used for interval mode
    
    # Determine GUI mode from config and CLI arguments
    gui_mode = determine_gui_mode(config_data, cli_gui=gui if gui is not None else None)
    logger.info(f"GUI mode determined: {gui_mode.value}")

    if autoclose:
        # Disable looping for all sources so that video files actually finish
        sources = config_data.get("pipeline", {}).get("sources", [])
        for source in sources:
            source["loop_play"] = False
        # Ensure controller section exists and propagate autoclose flag
        controller_cfg["autoclose"] = True
        logger.info("Auto-close enabled (loop_play disabled, controller.autoclose=True)")

    # No CLI recording overrides; use only config

    # Step 1: Initialize core (Controller) - always done first
    logger.info("Step 1: Initializing core system (Controller)")
    
    # Create initialization thread for core components
    init_thread = ControllerInitThread(config_data)
    initialization_result = {'controller': None, 'error': None, 'completed': False}
    
    # Progress callback for initialization
    progress_callback = None
    progress_window = None
    qt_app = None
    
    # Step 2: Initialize GUI based on mode
    gui_manager = None
    qt_app = None
    
    if gui_mode == GUIMode.HEADLESS:
        logger.info("Headless mode: skipping GUI initialization")
        # No GUI components, no QApplication
    else:
        # Create GUI manager for HIDDEN or VISIBLE modes
        logger.info(f"Step 2: Initializing GUI (mode: {gui_mode.value})")
        gui_manager = GUIManager(gui_mode, config_data, logger=logger)
        qt_app = gui_manager.create_gui_application()
        
        # Show progress window only in VISIBLE mode
        if gui_mode == GUIMode.VISIBLE:
            progress_window = StartupProgressWindow()
            progress_window.show()
            progress_window.raise_()
            qt_app.processEvents()
            logger.info("Progress window displayed")
            
            def on_progress_updated(value, stage_text):
                if progress_window:
                    progress_window.update_progress(value, stage_text)
                    qt_app.processEvents()
            progress_callback = on_progress_updated
    
    # Set progress callback for controller initialization
    if progress_callback:
        init_thread.progress_updated.connect(progress_callback)
    
    main_window_created = {'done': False}
    
    def on_initialization_complete(controller_instance):
        initialization_result['controller'] = controller_instance
        initialization_result['completed'] = True
        logger.info("Controller initialization completed")
        
        # Step 3: Connect GUI to controller (if GUI mode)
        if gui_mode != GUIMode.HEADLESS and gui_manager:
            try:
                logger.info("Step 3: Connecting GUI to controller")
                _connect_gui_to_controller(
                    controller_instance, config_file_name, config_data, 
                    gui_manager, logger, progress_window
                )
                logger.info("GUI connected to controller successfully")
                main_window_created['done'] = True
            except Exception as e:
                logger.error(f"Error connecting GUI to controller: {e}", exc_info=True)
                main_window_created['done'] = True
                raise
        else:
            # Headless mode: just start controller
            logger.info("Starting controller (headless mode)")
            controller_instance.start()
            main_window_created['done'] = True
    
    def on_initialization_failed(error_message):
        initialization_result['error'] = error_message
        initialization_result['completed'] = True
        main_window_created['done'] = True
        if progress_window:
            progress_window.close()
        logger.error(f"Controller initialization failed: {error_message}")
        if gui_mode != GUIMode.HEADLESS:
            QMessageBox.critical(None, "Initialization Error", 
                               f"Failed to initialize application:\n\n{error_message}")
        sys.exit(1)
    
    init_thread.initialization_complete.connect(on_initialization_complete)
    init_thread.initialization_failed.connect(on_initialization_failed)
    
    # Start core initialization
    logger.info("Starting controller initialization in background thread")
    init_thread.start()
    
    # Wait for initialization to complete
    if qt_app:
        # GUI mode: process events while waiting
        while not initialization_result['completed'] or not main_window_created['done']:
            qt_app.processEvents()
    else:
        # Headless mode: just wait (no event loop)
        import time
        while not initialization_result['completed'] or not main_window_created['done']:
            time.sleep(0.01)
    
    # Check result
    if initialization_result['error']:
        logger.error("Controller initialization failed")
        sys.exit(1)
    
    controller_instance = initialization_result['controller']
    
    # Step 4: Start application event loop
    if gui_mode == GUIMode.HEADLESS:
        logger.info("Headless mode: no visible GUI, starting minimal event loop")
        # In headless mode, controller runs until sources end; we still need a minimal
        # Qt event loop for timers and signal delivery.
        try:
            import os as _os
            _os.environ.setdefault("QT_NO_GLIB", "1")
        except Exception:
            pass
        qt_app = QApplication.instance() or QApplication(sys.argv)

        # Periodically check controller state and quit the event loop when it stops.
        # This is required so that headless runs with --autoclose (or other stop
        # conditions) actually terminate and allow the CLI scheduler to start the
        # next run.
        if qt_app is not None and controller_instance is not None:
            def _check_controller_finished():
                try:
                    if not controller_instance.is_running():
                        logger.info("Controller stopped in headless mode, quitting Qt event loop")
                        # Check if restart is requested due to memory leak
                        restart_requested = controller_instance.get_restart_flag()
                        cli_launched = os.environ.get('EVILEYE_CLI_LAUNCHED') == '1'
                        
                        if restart_requested:
                            if cli_launched:
                                logger.info("Memory leak detected: restart requested, CLI will handle restart")
                                qt_app.quit()
                                # Exit code 2 signals CLI scheduler to restart immediately
                                sys.exit(2)
                            else:
                                logger.warning("Memory leak detected but restart impossible: not launched via CLI")
                                qt_app.quit()
                                sys.exit(1)
                        else:
                            qt_app.quit()
                except Exception as e:
                    try:
                        logger.error(f"Error while checking controller state in headless mode: {e}", exc_info=True)
                    except Exception:
                        # Avoid cascading failures during shutdown
                        pass

            timer = QTimer(qt_app)
            timer.setInterval(1000)  # check every second
            timer.timeout.connect(_check_controller_finished)
            timer.start()

        logger.info("Controller started in headless mode, entering Qt event loop...")
        ret = qt_app.exec()
    else:
        logger.info("Starting main application loop")
        ret = qt_app.exec()
        
        # Check if restart is requested due to memory leak (GUI mode)
        if controller_instance is not None:
            restart_requested = controller_instance.get_restart_flag()
            cli_launched = os.environ.get('EVILEYE_CLI_LAUNCHED') == '1'
            
            if restart_requested:
                if cli_launched:
                    logger.info("Memory leak detected: restart requested, CLI will handle restart")
                    # Exit code 2 signals CLI scheduler to restart immediately
                    sys.exit(2)
                else:
                    logger.warning("Memory leak detected but restart impossible: not launched via CLI")
                    sys.exit(1)
    
    logger.info(f"Application finished with code: {ret}")
    return ret


def _connect_gui_to_controller(controller_instance, config_file_name, config_data, 
                                gui_manager: GUIManager, logger, progress_window=None):
    """
    Connect GUI components to controller.
    
    This function creates GUI components БЕЗ controller and connects them to the controller
    through the visualization adapter after controller initialization.
    """
    try:
        # Update progress if available
        if progress_window:
            progress_window.update_progress(90, "Creating GUI components...")
        
        # Create MainWindow БЕЗ controller (все виджеты создаются пустыми)
        logger.info("Creating MainWindow without controller...")
        main_window = MainWindow(1600, 720)
        gui_manager.main_window = main_window
        
        if progress_window:
            progress_window.update_progress(92, "Initializing GUI components...")
        
        # Update controller's GUI flags
        controller_instance.show_main_gui = (gui_manager.mode == GUIMode.VISIBLE)
        controller_instance.show_journal = config_data.get('controller', {}).get('show_journal', False)
        
        # Устанавливаем данные в MainWindow из controller ПЕРЕД созданием визуализатора
        # Это нужно для создания labels, которые визуализатор будет использовать
        logger.info("Setting controller data in MainWindow...")
        main_window.set_controller(controller_instance, config_file_name, config_data)
        logger.info("Controller data set in MainWindow")
        
        # Initialize visualizer using legacy method (init_main_window)
        # Визуализатор создается ПОСЛЕ установки данных, чтобы labels уже существовали
        logger.info("Initializing visualizer through controller...")
        controller_instance.init_main_window(main_window, main_window.slots, main_window.signals)
        logger.info("Visualizer initialized through controller")
        
        if progress_window:
            progress_window.update_progress(95, "Starting controller components...")
        
        # Start controller
        controller_instance.start()
        
        if progress_window:
            progress_window.update_progress(99, "Finalizing startup...")
        
        # Show GUI based on mode
        if gui_manager.mode == GUIMode.VISIBLE:
            gui_manager.show_gui()
            # Open journal if needed (from config)
            controller_config = config_data.get('controller', {})
            if controller_config.get('show_journal', False):
                # Journal window is created lazily, so we need to check if it exists or can be created
                if hasattr(main_window, 'db_journal_win') and main_window.db_journal_win:
                    main_window.open_journal()
                elif hasattr(main_window, '_deferred_journal_creation'):
                    # Journal will be created on first open
                    main_window.open_journal()
        elif gui_manager.mode == GUIMode.HIDDEN:
            # Hidden mode: start flag monitor
            gui_manager._start_flag_monitor()
            logger.info("Hidden GUI mode: GUI will be shown when flag file is detected")
        
        if progress_window:
            progress_window.update_progress(100, "Ready!")
            progress_window.hide()
            progress_window.close()
            progress_window.deleteLater()
            qt_app = gui_manager.get_qt_application()
            if qt_app:
                qt_app.processEvents()
        
        logger.info("GUI connected to controller successfully")
        
    except Exception as e:
        logger.error(f"Error connecting GUI to controller: {e}", exc_info=True)
        if progress_window:
            try:
                progress_window.close()
            except:
                pass
        raise


def _create_main_window_and_start(controller_instance, config_file_name, config_data, qt_app, gui, logger, progress_window=None):
    """Создать главное окно и запустить контроллер"""
    try:
        # Обновляем прогресс при создании главного окна
        if progress_window:
            progress_window.update_progress(90, "Creating main window...")
            qt_app.processEvents()
        
        logger.info("Creating main window (no-show in headless mode)")
        try:
            # Create MainWindow БЕЗ controller (все виджеты создаются пустыми)
            a = MainWindow(1600, 720)
            logger.info("MainWindow instance created successfully")
        except Exception as e:
            logger.error(f"Error creating MainWindow: {e}", exc_info=True)
            raise
        
        logger.info("DEBUG: After MainWindow creation, before progress update")
        if progress_window:
            logger.info("DEBUG: Updating progress to 92%")
            try:
                progress_window.update_progress(92, "Initializing main window components...")
                logger.info("DEBUG: Progress updated to 92%")
            except Exception as e:
                logger.error(f"DEBUG: Error updating progress: {e}", exc_info=True)
        else:
            logger.info("DEBUG: progress_window is None")
        
        logger.info("DEBUG: MainWindow instance created, calling init_main_window...")
        controller_instance.init_main_window(a, a.slots, a.signals)
        logger.info(f"init_main_window completed, show_main_gui={controller_instance.show_main_gui}, gui={gui}")
        
        # Устанавливаем данные в MainWindow из controller
        logger.info("Setting controller data in MainWindow...")
        a.set_controller(controller_instance, config_file_name, config_data)
        logger.info("Controller data set in MainWindow")
        
        # Устанавливаем callback для обновления прогресса при запуске модулей
        logger.info("Setting up progress callback...")
        if progress_window:
            def start_progress_callback(value, stage_text):
                # Минимальный callback без блокирующих операций
                try:
                    if progress_window and hasattr(progress_window, 'progress_bar'):
                        # Используем значение напрямую, так как start() теперь использует 95-98%
                        progress_window.progress_bar.setValue(value)
                        if stage_text and hasattr(progress_window, 'stage_label'):
                            progress_window.stage_label.setText(stage_text)
                except Exception:
                    # Игнорируем ошибки в callback, чтобы не блокировать выполнение
                    pass
            controller_instance.progress_callback = start_progress_callback
            logger.info("Progress callback set up")
        else:
            # Убираем callback, если окно прогресса не доступно
            controller_instance.progress_callback = None
            logger.info("Progress callback disabled (no progress window)")

        logger.info("DEBUG: After callback setup")
        if progress_window:
            logger.info("DEBUG: progress_window exists, updating to 95%")
            try:
                progress_window.progress_bar.setValue(95)
                progress_window.stage_label.setText("Starting controller components...")
                logger.info("DEBUG: Progress updated to 95%")
            except Exception as e:
                logger.error(f"DEBUG: Error updating progress: {e}", exc_info=True)
        else:
            logger.info("DEBUG: progress_window is None")

        logger.info("DEBUG: About to call controller.start()...")
        try:
            controller_instance.start()
            logger.info("Controller started successfully")
        except Exception as e:
            logger.error(f"Error starting controller: {e}", exc_info=True)
            raise
        
        if progress_window:
            progress_window.update_progress(99, "Finalizing startup...")
            qt_app.processEvents()
        
        # Закрываем прогресс-бар после завершения запуска всех модулей
        if progress_window:
            progress_window.update_progress(100, "Ready!")
            qt_app.processEvents()
            logger.info("Closing progress window")
            try:
                if progress_window.isVisible():
                    progress_window.hide()  # Скрываем окно
                progress_window.close()  # Закрываем окно
                progress_window.setParent(None)  # Удаляем ссылку на родителя
                progress_window.deleteLater()  # Планируем удаление окна
                qt_app.processEvents()  # Обновляем GUI, чтобы окно закрылось
                logger.info("Progress window closed")
            except Exception as e:
                logger.warning(f"Error closing progress window: {e}")
        
        # Показываем главное окно ПОСЛЕ завершения start() и закрытия прогресс-бара
        if gui and controller_instance.show_main_gui:
            logger.info("Showing main window")
            a.show()
            a.raise_()  # Поднимаем окно на передний план
            a.activateWindow()  # Активируем окно
            qt_app.processEvents()  # Обновляем GUI
            logger.info("Main window displayed and activated")
            if controller_instance.show_journal:
                a.open_journal()
                logger.info("Journal opened")
        else:
            logger.info(f"Main window not shown: gui={gui}, show_main_gui={controller_instance.show_main_gui}")
    except Exception as e:
        logger.error(f"Error in _create_main_window_and_start: {e}", exc_info=True)
        if progress_window:
            try:
                progress_window.close()
            except:
                pass
        raise





from typing import Optional
from .pipeline_manager import PipelineManager

""" Module for accessing the pipeline manager, singleton """

_manager: Optional[PipelineManager] = None


def get_manager() -> PipelineManager:
    global _manager
    if _manager is None:
        _manager = PipelineManager()
    return _manager



import copy
import json
import os.path
try:
    from PyQt6 import QtGui
    from PyQt6.QtWidgets import (
        QWidget, QLabel, QVBoxLayout, QHBoxLayout, QLineEdit, QScrollArea,
        QSizePolicy, QToolBar, QComboBox, QFormLayout, QSpacerItem,
        QMenu, QMainWindow, QApplication, QCheckBox, QPushButton, QTabWidget
    )
    from PyQt6.QtGui import QIcon
    from PyQt6.QtGui import QAction
    from PyQt6.QtCore import pyqtSignal, pyqtSlot, Qt
    pyqt_version = 6
except ImportError:
    from PyQt5 import QtGui
    from PyQt5.QtWidgets import (
        QWidget, QLabel, QVBoxLayout, QHBoxLayout, QLineEdit, QScrollArea,
        QSizePolicy, QToolBar, QComboBox, QFormLayout, QSpacerItem,
        QMenu, QMainWindow, QApplication, QCheckBox, QPushButton, QTabWidget
    )
    from PyQt5.QtGui import QIcon
    from PyQt5.QtWidgets import QAction
    from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt
    pyqt_version = 5
from evileye.utils import utils
import sys
from evileye.capture.video_capture_base import CaptureDeviceType
from evileye.capture import VideoCaptureOpencv
from evileye.visualization_modules.configurer import parameters_processing


class TrackerWidget(QWidget):
    def __init__(self, params):
        super().__init__()

        self.params = params
        self.proj_root = utils.get_project_root()
        self.track_layout = None

        self.line_edit_param = {}  # Словарь для сопоставления полей интерфейса с полями json-файла

        self.horizontal_layout = QHBoxLayout()
        self.horizontal_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._setup_tracker_layout(self.params)
        self.setLayout(self.horizontal_layout)

    def _setup_tracker_layout(self, params):
        self.track_layout = self._setup_tracker_form(params)
        self.horizontal_layout.addLayout(self.track_layout)

    def _setup_tracker_form(self, params) -> QFormLayout:
        layout = QFormLayout()

        name = QLabel('Tracker Parameters')
        layout.addWidget(name)

        src_ids = QLineEdit()
        src_ids.setText(str(params['source_ids']))
        layout.addRow('Sources ids', src_ids)
        self.line_edit_param['source_ids'] = src_ids

        fps = QLineEdit()
        fps.setText(str(params['fps']))
        layout.addRow('FPS', fps)
        self.line_edit_param['fps'] = fps

        tracker_onnx = QLineEdit()
        tracker_onnx.setText(str(params.get('tracker_onnx', None)))
        layout.addRow('Tracker ONNX', tracker_onnx)
        self.line_edit_param['tracker_onnx'] = tracker_onnx

        botsort_params = params['botsort_cfg']
        self._setup_botsort_params(layout, botsort_params)

        widgets = (layout.itemAt(i).widget() for i in range(layout.count()))
        for widget in widgets:
            widget.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
            widget.setMinimumWidth(200)
        return layout

    def _setup_botsort_params(self, layout, params):
        appearance_thresh = QLineEdit()
        appearance_thresh.setText(str(params['appearance_thresh']))
        layout.addRow('Appearance threshold', appearance_thresh)
        self.line_edit_param['appearance_thresh'] = appearance_thresh

        gmc_method = QLineEdit()
        layout.addRow('GMC method', gmc_method)
        gmc_method.setText(str(params['gmc_method']))
        self.line_edit_param['gmc_method'] = gmc_method

        match_thresh = QLineEdit()
        layout.addRow('Match threshold', match_thresh)
        match_thresh.setText(str(params['match_thresh']))
        self.line_edit_param['match_thresh'] = match_thresh

        new_track_thresh = QLineEdit()
        layout.addRow('New track threshold', new_track_thresh)
        new_track_thresh.setText(str(params['new_track_thresh']))
        self.line_edit_param['new_track_thresh'] = new_track_thresh

        proximity_thresh = QLineEdit()
        layout.addRow('Proximity threshold', proximity_thresh)
        proximity_thresh.setText(str(params['proximity_thresh']))
        self.line_edit_param['proximity_thresh'] = proximity_thresh

        track_buffer = QLineEdit()
        layout.addRow('Track buffer', track_buffer)
        track_buffer.setText(str(params['track_buffer']))
        self.line_edit_param['track_buffer'] = track_buffer

        track_high_thresh = QLineEdit()
        layout.addRow('High threshold', track_high_thresh)
        track_high_thresh.setText(str(params['track_high_thresh']))
        self.line_edit_param['track_high_thresh'] = track_high_thresh

        track_low_thresh = QLineEdit()
        layout.addRow('Low threshold', track_low_thresh)
        track_low_thresh.setText(str(params['track_low_thresh']))
        self.line_edit_param['track_low_thresh'] = track_low_thresh

        tracker_type = QLineEdit()
        layout.addRow('Tracker type', tracker_type)
        tracker_type.setText(str(params['tracker_type']))
        self.line_edit_param['tracker_type'] = tracker_type

        with_reid = QCheckBox()
        layout.addRow('With ReId', with_reid)
        self.line_edit_param['with_reid'] = with_reid
        flag = params.get('with_reid', None)
        if flag == '' or flag is None:
            flag = False
        with_reid.setChecked(flag)
        # with_reid.setEnabled(False)
        # with_reid.checkStateChanged.connect(self._display_reid_params)

    # @pyqtSlot()
    # def _display_reid_params(self):
    #     # Индекс умножается на два из-за наличия spacers
    #     onnx_widgets = (self.track_layout.itemAt(i).widget()
    #                     for i in range(self.track_layout.count() - 2, self.track_layout.count()))
    #     if self.sender().isChecked():
    #         for widget in onnx_widgets:
    #             widget.setVisible(True)
    #     else:
    #         for widget in onnx_widgets:
    #             widget.setVisible(False)

    def get_form(self) -> QFormLayout:
        return self.track_layout

    def get_params(self):
        return self.line_edit_param.get('trackers', None)

    def get_dict(self):
        res_dict = self._create_dict()
        return res_dict

    def _create_dict(self):
        src_params = {}

        widget = self.line_edit_param['source_ids']
        src_params['source_ids'] = parameters_processing.process_numeric_lists(widget.text())

        widget = self.line_edit_param['fps']
        src_params['fps'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['tracker_onnx']
        if widget.text() == 'None':
            src_params['inference_size'] = None
        else:
            src_params['inference_size'] = widget.text()

        botsort_params = {}

        widget = self.line_edit_param['appearance_thresh']
        botsort_params['appearance_thresh'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['gmc_method']
        botsort_params['gmc_method'] = widget.text()

        widget = self.line_edit_param['match_thresh']
        botsort_params['match_thresh'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['new_track_thresh']
        botsort_params['new_track_thresh'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['proximity_thresh']
        botsort_params['proximity_thresh'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['track_buffer']
        botsort_params['track_buffer'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['track_high_thresh']
        botsort_params['track_high_thresh'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['track_low_thresh']
        botsort_params['track_low_thresh'] = parameters_processing.process_numeric_types(widget.text())

        widget = self.line_edit_param['tracker_type']
        botsort_params['tracker_type'] = widget.text()

        widget = self.line_edit_param['with_reid']
        botsort_params['with_reid'] = True if widget.isChecked() else False

        src_params['botsort_cfg'] = botsort_params
        return src_params



from fastapi import APIRouter, HTTPException, Query, WebSocket, WebSocketDisconnect
from typing import List, Optional, Dict, Any
from pydantic import BaseModel

from evileye.core.logger import get_module_logger
from evileye.api.core.manager_access import get_manager
from evileye.api.core.pipeline_manager import PipelineState
import asyncio

logger = get_module_logger("api.events")

router = APIRouter(prefix="/api/v1", tags=["events"])


class ObjectBbox(BaseModel):
    x1: float
    y1: float
    x2: float
    y2: float


class ObjectInfo(BaseModel):
    object_id: int
    source_id: int
    class_id: int
    class_name: Optional[str] = None
    confidence: float
    bbox: ObjectBbox
    track_id: Optional[int] = None
    timestamp: Optional[str] = None
    properties: Dict[str, Any] = {}


class EventInfo(BaseModel):
    event_id: Optional[int] = None
    event_type: str
    source_id: Optional[int] = None
    object_id: Optional[int] = None
    timestamp: str
    metadata: Dict[str, Any] = {}


@router.get("/pipelines/{rid}/objects")
async def get_objects(
    rid: int,
    object_type: str = Query("active", description="Object type: 'active', 'lost', 'all'")
) -> List[ObjectInfo]:
    """
    Get tracked objects from the pipeline.
    
    Parameters:
    - pid: Pipeline ID
    - object_type: Type of objects to retrieve ('active', 'lost', 'all')
    """
    logger.info(f"GET /pipelines/{rid}/objects: object_type={object_type}")
    try:
        pipeline_info = get_manager().describe(rid)
        if pipeline_info["state"] != PipelineState.RUNNING:
            logger.warning(f"Pipeline '{rid}' is not running (state: {pipeline_info['state']})")
            raise HTTPException(status_code=400, detail=f"Pipeline '{rid}' is not running")
    except KeyError:
        logger.error(f"Pipeline '{rid}' not found")
        raise HTTPException(status_code=404, detail=f"Pipeline '{rid}' not found")
    
    try:
        runner = get_manager()._get_runner(rid)
        if runner.controller is None:
            logger.warning(f"Controller is None for pipeline '{rid}'")
            return []
        
        obj_handler = runner.controller.obj_handler
        if obj_handler is None:
            logger.warning(f"Object handler is None for pipeline '{pid}'")
            return []
        
        objects = []
        sources = runner.controller.pipeline.get_sources() if hasattr(runner.controller, 'pipeline') else []
        source_ids = [src.get_id() for src in sources] if sources else [0]
        
        for source_id in source_ids:
            obj_list = obj_handler.get(object_type, source_id)
            if not obj_list or not hasattr(obj_list, 'objects'):
                continue
            
            for obj in obj_list.objects:
                track = getattr(obj, 'track', None)
                bbox = getattr(track, 'bounding_box', [0, 0, 0, 0]) if track else [0, 0, 0, 0]
                confidence = getattr(track, 'confidence', 0.0) if track else 0.0
                track_id = getattr(track, 'track_id', None) if track else None
                
                obj_data = {
                    "object_id": getattr(obj, 'object_id', 0),
                    "source_id": getattr(obj, 'source_id', source_id),
                    "class_id": getattr(obj, 'class_id', 0),
                    "class_name": getattr(obj, 'class_name', None),
                    "confidence": confidence,
                    "bbox": {
                        "x1": bbox[0] if len(bbox) >= 1 else 0,
                        "y1": bbox[1] if len(bbox) >= 2 else 0,
                        "x2": bbox[2] if len(bbox) >= 3 else 0,
                        "y2": bbox[3] if len(bbox) >= 4 else 0
                    },
                    "track_id": track_id,
                    "timestamp": str(getattr(obj, 'time_stamp', '')),
                    "properties": getattr(obj, 'properties', {})
                }
                objects.append(ObjectInfo(**obj_data))
        
        logger.info(f"Retrieved {len(objects)} objects for pipeline '{rid}'")
        return objects
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Failed to retrieve objects for pipeline '{rid}': {e}")
        raise HTTPException(status_code=500, detail=f"Failed to retrieve objects: {str(e)}")


@router.get("/pipelines/{rid}/events")
async def get_events(
    rid: int,
    event_type: Optional[str] = Query(None, description="Filter by event type")
) -> List[EventInfo]:
    """
    Get events from the pipeline.
    
    Parameters:
    - pid: Pipeline ID
    - event_type: Optional filter by event type
    """
    logger.info(f"GET /pipelines/{rid}/events: event_type={event_type}")
    try:
        pipeline_info = get_manager().describe(rid)
        if pipeline_info["state"] != PipelineState.RUNNING:
            logger.warning(f"Pipeline '{rid}' is not running for WebSocket (state: {pipeline_info['state']})")
            raise HTTPException(status_code=400, detail=f"Pipeline '{rid}' is not running")
    except KeyError:
        logger.error(f"Pipeline '{rid}' not found for WebSocket")
        raise HTTPException(status_code=404, detail=f"Pipeline '{rid}' not found")
    
    try:
        runner = get_manager()._get_runner(rid)
        if runner.controller is None:
            logger.warning(f"Controller is None for pipeline '{rid}'")
            return []
        
        events_detector = runner.controller.events_detectors_controller
        if events_detector is None:
            logger.warning(f"Events detector is None for pipeline '{rid}'")
            return []
        
        events_list = []
        try:
            events_dict = events_detector.events_detectors if hasattr(events_detector, 'events_detectors') else events_detector.get()
            
            if events_dict:
                for detector_name, events in events_dict.items():
                    if events and isinstance(events, list):
                        for event in events:
                            if hasattr(event, 'get_name'):
                                event_type_name = event.get_name()
                                if event_type and event_type != event_type_name:
                                    continue
                                
                                event_data = {
                                    "event_type": event_type_name,
                                    "source_id": getattr(event, 'source_id', None),
                                    "object_id": getattr(event, 'object_id', None),
                                    "timestamp": str(getattr(event, 'timestamp', '')),
                                    "metadata": {}
                                }
                                events_list.append(EventInfo(**event_data))
        except Exception as e:
            logger.debug(f"Error accessing events: {e}")
        
        logger.info(f"Retrieved {len(events_list)} events for pipeline '{rid}'")
        return events_list
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Failed to retrieve events for pipeline '{rid}': {e}")
        raise HTTPException(status_code=500, detail=f"Failed to retrieve events: {str(e)}")


@router.get("/pipelines/{rid}/metadata")
async def get_metadata(rid: int) -> Dict[str, Any]:
    """
    Get pipeline metadata including statistics and configuration.
    
    Parameters:
    - pid: Pipeline ID
    """
    logger.info(f"GET /pipelines/{rid}/metadata")
    try:
        pipeline_info = get_manager().describe(rid)
    except KeyError:
        logger.error(f"Pipeline '{rid}' not found")
        raise HTTPException(status_code=404, detail=f"Pipeline '{rid}' not found")
    
    try:
        runner = get_manager()._get_runner(rid)
        metadata = {
            "pipeline_id": rid,
            "pipeline_name": pipeline_info.get("name"),
            "state": pipeline_info.get("state"),
            "statistics": {}
        }
        
        if runner.controller is not None:
            obj_handler = runner.controller.obj_handler
            if obj_handler:
                active_count = len(obj_handler.active_objs.objects) if hasattr(obj_handler, 'active_objs') else 0
                lost_count = len(obj_handler.lost_objs.objects) if hasattr(obj_handler, 'lost_objs') else 0
                metadata["statistics"] = {
                    "active_objects": active_count,
                    "lost_objects": lost_count
                }
            
            if hasattr(runner.controller, 'pipeline') and runner.controller.pipeline:
                sources = runner.controller.pipeline.get_sources()
                metadata["sources"] = [
                    {
                        "source_id": src.get_id(),
                        "source_name": getattr(src, 'source_name', f"Source-{src.get_id()}")
                    }
                    for src in sources
                ]
        
        logger.info(f"Retrieved metadata for pipeline '{rid}'")
        return metadata
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Failed to retrieve metadata for pipeline '{rid}': {e}")
        raise HTTPException(status_code=500, detail=f"Failed to retrieve metadata: {str(e)}")


def _get_objects_data(runner, object_type: str = "active") -> List[dict]:
    """Helper function to extract objects data from pipeline."""
    if runner.controller is None:
        return []
    
    obj_handler = runner.controller.obj_handler
    if obj_handler is None:
        return []
    
    objects = []
    sources = runner.controller.pipeline.get_sources() if hasattr(runner.controller, 'pipeline') else []
    source_ids = [src.get_id() for src in sources] if sources else [0]
    
    for source_id in source_ids:
        obj_list = obj_handler.get(object_type, source_id)
        if not obj_list or not hasattr(obj_list, 'objects'):
            continue
        
        for obj in obj_list.objects:
            track = getattr(obj, 'track', None)
            bbox = getattr(track, 'bounding_box', [0, 0, 0, 0]) if track else [0, 0, 0, 0]
            confidence = getattr(track, 'confidence', 0.0) if track else 0.0
            track_id = getattr(track, 'track_id', None) if track else None
            
            obj_data = {
                "object_id": getattr(obj, 'object_id', 0),
                "source_id": getattr(obj, 'source_id', source_id),
                "class_id": getattr(obj, 'class_id', 0),
                "class_name": getattr(obj, 'class_name', None),
                "confidence": confidence,
                "bbox": {
                    "x1": float(bbox[0]) if len(bbox) >= 1 else 0,
                    "y1": float(bbox[1]) if len(bbox) >= 2 else 0,
                    "x2": float(bbox[2]) if len(bbox) >= 3 else 0,
                    "y2": float(bbox[3]) if len(bbox) >= 4 else 0
                },
                "track_id": track_id,
                "timestamp": str(getattr(obj, 'time_stamp', '')),
                "properties": getattr(obj, 'properties', {})
            }
            objects.append(obj_data)
    
    return objects


@router.websocket("/pipelines/{rid}/stream/data")
async def websocket_data_stream(
    websocket: WebSocket,
    rid: int,
    update_interval: float = 0.1,
    include_objects: bool = True,
    include_events: bool = True
):
    """
    WebSocket endpoint for real-time objects and events streaming.
    
    Parameters:
    - pid: Pipeline ID
    - update_interval: Update interval in seconds (default: 0.1)
    - include_objects: Include objects in stream (default: true)
    - include_events: Include events in stream (default: true)
    """
    logger.info(f"WebSocket connection established for pipeline '{rid}'")
    await websocket.accept()
    
    try:
        pipeline_info = get_manager().describe(rid)
        if pipeline_info["state"] != PipelineState.RUNNING:
            logger.warning(f"Pipeline '{rid}' is not running for WebSocket (state: {pipeline_info['state']})")
            await websocket.send_json({"error": f"Pipeline '{pid}' is not running"})
            return
    except KeyError:
        logger.error(f"Pipeline '{rid}' not found for WebSocket")
        await websocket.send_json({"error": f"Pipeline '{pid}' not found"})
        return
    
    try:
        runner = get_manager()._get_runner(rid)
        logger.info(f"WebSocket stream started for pipeline '{rid}', update_interval={update_interval}s")
        
        while True:
            data = {}
            
            if include_objects:
                objects = _get_objects_data(runner, "active")
                data["objects"] = objects
            
            if include_events:
                events_detector = runner.controller.events_detectors_controller if runner.controller else None
                if events_detector:
                    try:
                        events_dict = events_detector.events_detectors if hasattr(events_detector, 'events_detectors') else events_detector.get()
                        events_list = []
                        if events_dict:
                            for detector_name, events in events_dict.items():
                                if events and isinstance(events, list):
                                    for event in events:
                                        if hasattr(event, 'get_name'):
                                            events_list.append({
                                                "event_type": event.get_name(),
                                                "source_id": getattr(event, 'source_id', None),
                                                "object_id": getattr(event, 'object_id', None),
                                                "timestamp": str(getattr(event, 'timestamp', ''))
                                            })
                        data["events"] = events_list
                    except Exception as e:
                        logger.debug(f"Error accessing events in WebSocket: {e}")
                        data["events"] = []
            
            data["metadata"] = {
                "pipeline_id": rid,
                "timestamp": str(asyncio.get_event_loop().time())
            }
            
            await websocket.send_json(data)
            await asyncio.sleep(update_interval)
            
    except WebSocketDisconnect:
        logger.info(f"WebSocket disconnected for pipeline '{rid}'")
        pass
    except Exception as e:
        logger.error(f"WebSocket error for pipeline '{rid}': {e}")
        await websocket.send_json({"error": str(e)})
